from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import collections
from functools import partial
from itertools import product

import numpy as np
import sklearn.metrics
from sklearn.utils.multiclass import type_of_target
from smac.utils.constants import MAXINT

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
    REGRESSION_TASKS,
    TASK_TYPES,
)
from autosklearn.data.target_validator import SUPPORTED_XDATA_TYPES

from .util import sanitize_array


class Scorer(object, metaclass=ABCMeta):
    def __init__(
        self,
        name: str,
        score_func: Callable,
        optimum: float,
        worst_possible_result: float,
        sign: float,
        kwargs: Any,
        needs_X: bool = False,
    ) -> None:
        self.name = name
        self._kwargs = kwargs
        self._score_func = score_func
        self._optimum = optimum
        self._needs_X = needs_X
        self._worst_possible_result = worst_possible_result
        self._sign = sign

    @abstractmethod
    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        *,
        X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
        sample_weight: Optional[List[float]] = None,
    ) -> float:
        pass

    def __repr__(self) -> str:
        return self.name


class _PredictScorer(Scorer):
    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        *,
        X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
        sample_weight: Optional[List[float]] = None,
    ) -> float:
        """Evaluate predicted target values for X relative to y_true.

        Parameters
        ----------
        y_true : array-like
            Gold standard target values for X.

        y_pred : array-like, [n_samples x n_classes]
            Model predictions

        X_data : array-like [n_samples x n_features]
            X data used to obtain the predictions: each row x_j corresponds to the input
             used to obtain predictions y_j

        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        type_true = type_of_target(y_true)
        if (
            type_true == "binary"
            and type_of_target(y_pred) == "continuous"
            and len(y_pred.shape) == 1
        ):
            # For a pred scorer, no threshold, nor probability is required
            # If y_true is binary, and y_pred is continuous
            # it means that a rounding is necessary to obtain the binary class
            y_pred = np.around(y_pred, decimals=0)
        elif (
            len(y_pred.shape) == 1 or y_pred.shape[1] == 1 or type_true == "continuous"
        ):
            # must be regression, all other task types would return at least
            # two probabilities
            pass
        elif type_true in ["binary", "multiclass"]:
            y_pred = np.argmax(y_pred, axis=1)
        elif type_true == "multilabel-indicator":
            y_pred[y_pred > 0.5] = 1.0
            y_pred[y_pred <= 0.5] = 0.0
        elif type_true == "continuous-multioutput":
            pass
        else:
            raise ValueError(type_true)

        scorer_kwargs = {}  # type: Dict[str, Union[List[float], np.ndarray]]
        if sample_weight is not None:
            scorer_kwargs["sample_weight"] = sample_weight
        if self._needs_X is True:
            scorer_kwargs["X_data"] = X_data

        return self._sign * self._score_func(
            y_true, y_pred, **scorer_kwargs, **self._kwargs
        )


class _ProbaScorer(Scorer):
    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        *,
        X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
        sample_weight: Optional[List[float]] = None,
    ) -> float:
        """Evaluate predicted probabilities for X relative to y_true.
        Parameters
        ----------
        y_true : array-like
            Gold standard target values for X. These must be class labels,
            not probabilities.

        y_pred : array-like, [n_samples x n_classes]
            Model predictions

        X_data : array-like [n_samples x n_features]
            X data used to obtain the predictions: each row x_j corresponds to the input
             used to obtain predictions y_j

        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """

        if self._score_func is sklearn.metrics.log_loss:
            n_labels_pred = np.array(y_pred).reshape((len(y_pred), -1)).shape[1]
            n_labels_test = len(np.unique(y_true))
            if n_labels_pred != n_labels_test:
                labels = list(range(n_labels_pred))
                if sample_weight is not None:
                    return self._sign * self._score_func(
                        y_true,
                        y_pred,
                        sample_weight=sample_weight,
                        labels=labels,
                        **self._kwargs,
                    )
                else:
                    return self._sign * self._score_func(
                        y_true, y_pred, labels=labels, **self._kwargs
                    )

        scorer_kwargs = {}  # type: Dict[str, Union[List[float], np.ndarray]]
        if sample_weight is not None:
            scorer_kwargs["sample_weight"] = sample_weight
        if self._needs_X is True:
            scorer_kwargs["X_data"] = X_data

        return self._sign * self._score_func(
            y_true, y_pred, **scorer_kwargs, **self._kwargs
        )


class _ThresholdScorer(Scorer):
    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        *,
        X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
        sample_weight: Optional[List[float]] = None,
    ) -> float:
        """Evaluate decision function output for X relative to y_true.
        Parameters
        ----------
        y_true : array-like
            Gold standard target values for X. These must be class labels,
            not probabilities.

        y_pred : array-like, [n_samples x n_classes]
            Model predictions

        X_data : array-like [n_samples x n_features]
            X data used to obtain the predictions: each row x_j corresponds to the input
             used to obtain predictions y_j

        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        y_type = type_of_target(y_true)
        if y_type not in ("binary", "multilabel-indicator"):
            raise ValueError("{0} format is not supported".format(y_type))

        if y_type == "binary":
            if y_pred.ndim > 1:
                y_pred = y_pred[:, 1]
        elif isinstance(y_pred, list):
            y_pred = np.vstack([p[:, -1] for p in y_pred]).T

        scorer_kwargs = {}  # type: Dict[str, Union[List[float], np.ndarray]]
        if sample_weight is not None:
            scorer_kwargs["sample_weight"] = sample_weight
        if self._needs_X is True:
            scorer_kwargs["X_data"] = X_data

        return self._sign * self._score_func(
            y_true, y_pred, **scorer_kwargs, **self._kwargs
        )


def make_scorer(
    name: str,
    score_func: Callable,
    *,
    optimum: float = 1.0,
    worst_possible_result: float = 0.0,
    greater_is_better: bool = True,
    needs_proba: bool = False,
    needs_threshold: bool = False,
    needs_X: bool = False,
    **kwargs: Any,
) -> Scorer:
    """Make a scorer from a performance metric or loss function.

    Factory inspired by scikit-learn which wraps scikit-learn scoring functions
    to be used in auto-sklearn.

    Parameters
    ----------
    name: str
        Descriptive name of the metric

    score_func : callable
        Score function (or loss function) with signature
        ``score_func(y, y_pred, **kwargs)``.

    optimum : int or float, default=1
        The best score achievable by the score function, i.e. maximum in case of
        scorer function and minimum in case of loss function.

    worst_possible_result : int of float, default=0
        The worst score achievable by the score function, i.e. minimum in case of
        scorer function and maximum in case of loss function.

    greater_is_better : boolean, default=True
        Whether score_func is a score function (default), meaning high is good,
        or a loss function, meaning low is good. In the latter case, the
        scorer object will sign-flip the outcome of the score_func.

    needs_proba : boolean, default=False
        Whether score_func requires predict_proba to get probability estimates
        out of a classifier.

    needs_threshold : boolean, default=False
        Whether score_func takes a continuous decision certainty.
        This only works for binary classification.

    needs_X : boolean, default=False
        Whether score_func requires X in __call__ to compute a metric.

    **kwargs : additional arguments
        Additional parameters to be passed to score_func.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score; greater is better or set
        greater_is_better to False.
    """
    sign = 1 if greater_is_better else -1
    if needs_proba and needs_threshold:
        raise ValueError(
            "Set either needs_proba or needs_threshold to True, but not both."
        )

    cls = None  # type: Any
    if needs_proba:
        cls = _ProbaScorer
    elif needs_threshold:
        cls = _ThresholdScorer
    else:
        cls = _PredictScorer
    return cls(
        name, score_func, optimum, worst_possible_result, sign, kwargs, needs_X=needs_X
    )


# Standard regression scores
mean_absolute_error = make_scorer(
    "mean_absolute_error",
    sklearn.metrics.mean_absolute_error,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
)
mean_squared_error = make_scorer(
    "mean_squared_error",
    sklearn.metrics.mean_squared_error,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
    squared=True,
)
root_mean_squared_error = make_scorer(
    "root_mean_squared_error",
    sklearn.metrics.mean_squared_error,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
    squared=False,
)
mean_squared_log_error = make_scorer(
    "mean_squared_log_error",
    sklearn.metrics.mean_squared_log_error,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
)
median_absolute_error = make_scorer(
    "median_absolute_error",
    sklearn.metrics.median_absolute_error,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
)

r2 = make_scorer("r2", sklearn.metrics.r2_score)

# Standard Classification Scores
accuracy = make_scorer("accuracy", sklearn.metrics.accuracy_score)
balanced_accuracy = make_scorer(
    "balanced_accuracy", sklearn.metrics.balanced_accuracy_score
)

# Score functions that need decision values
roc_auc = make_scorer(
    "roc_auc",
    sklearn.metrics.roc_auc_score,
    greater_is_better=True,
    needs_threshold=True,
)
average_precision = make_scorer(
    "average_precision", sklearn.metrics.average_precision_score, needs_threshold=True
)

# NOTE: zero_division
#
#   Specified as the explicit default, see sklearn docs:
#   https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html#sklearn-metrics-precision-score
precision = make_scorer(
    "precision", partial(sklearn.metrics.precision_score, zero_division=0)
)
recall = make_scorer("recall", partial(sklearn.metrics.recall_score, zero_division=0))
f1 = make_scorer("f1", partial(sklearn.metrics.f1_score, zero_division=0))

# Score function for probabilistic classification
log_loss = make_scorer(
    "log_loss",
    sklearn.metrics.log_loss,
    optimum=0,
    worst_possible_result=MAXINT,
    greater_is_better=False,
    needs_proba=True,
)
# TODO what about mathews correlation coefficient etc?


REGRESSION_METRICS = {
    scorer.name: scorer
    for scorer in [
        mean_absolute_error,
        mean_squared_error,
        root_mean_squared_error,
        mean_squared_log_error,
        median_absolute_error,
        r2,
    ]
}

CLASSIFICATION_METRICS = {
    scorer.name: scorer
    for scorer in [accuracy, balanced_accuracy, roc_auc, average_precision, log_loss]
}

# NOTE: zero_division
#
#   Specified as the explicit default, see sklearn docs:
#   https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html#sklearn-metrics-precision-score
for (base_name, sklearn_metric), average in product(
    [
        ("precision", sklearn.metrics.precision_score),
        ("recall", sklearn.metrics.recall_score),
        ("f1", sklearn.metrics.f1_score),
    ],
    ["macro", "micro", "samples", "weighted"],
):
    name = f"{base_name}_{average}"
    scorer = make_scorer(
        name, partial(sklearn_metric, pos_label=None, average=average, zero_division=0)
    )
    globals()[name] = scorer  # Adds scorer to the module scope
    CLASSIFICATION_METRICS[name] = scorer


def _validate_metrics(
    metrics: Sequence[Scorer],
    scoring_functions: Optional[List[Scorer]] = None,
) -> None:
    """
    Validate metrics given to Auto-sklearn. Raises an Exception in case of a problem.

    metrics: Sequence[Scorer]
        A list of objects that hosts a function to calculate how good the
        prediction is according to the solution.
    scoring_functions: Optional[List[Scorer]]
        A list of metrics to calculate multiple losses
    """

    to_score = list(metrics)
    if scoring_functions:
        to_score.extend(scoring_functions)

    if len(metrics) == 0:
        raise ValueError("Number of metrics to compute must be greater than zero.")

    metric_counter = collections.Counter(to_score)
    metric_names_counter = collections.Counter(metric.name for metric in to_score)
    if len(metric_counter) != len(metric_names_counter):
        raise ValueError(
            "Error in metrics passed to Auto-sklearn. A metric name was used "
            "multiple times for different metrics!"
        )


def calculate_scores(
    solution: np.ndarray,
    prediction: np.ndarray,
    task_type: int,
    metrics: Sequence[Scorer],
    *,
    X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
    scoring_functions: Optional[List[Scorer]] = None,
) -> Dict[str, float]:
    """
    Returns the scores (a magnitude that allows casting the
    optimization problem as a maximization one) for the
    given Auto-Sklearn Scorer objects.

    Parameters
    ----------
    solution: np.ndarray
        The ground truth of the targets
    prediction: np.ndarray
        The best estimate from the model, of the given targets
    task_type: int
        To understand if the problem task is classification
        or regression
    metrics: Sequence[Scorer]
        A list of objects that hosts a function to calculate how good the
        prediction is according to the solution.
    X_data : array-like [n_samples x n_features]
        X data used to obtain the predictions
    scoring_functions: List[Scorer]
        A list of metrics to calculate multiple losses
    Returns
    -------
    Dict[str, float]
    """
    if task_type not in TASK_TYPES:
        raise NotImplementedError(task_type)

    _validate_metrics(metrics=metrics, scoring_functions=scoring_functions)

    to_score = list(metrics)
    if scoring_functions:
        to_score.extend(scoring_functions)

    score_dict = dict()
    if task_type in REGRESSION_TASKS:
        for metric_ in to_score:

            try:
                score_dict[metric_.name] = _compute_single_scorer(
                    metric=metric_,
                    prediction=prediction,
                    solution=solution,
                    task_type=task_type,
                    X_data=X_data,
                )
            except ValueError as e:
                print(e, e.args[0])
                if (
                    e.args[0] == "Mean Squared Logarithmic Error cannot be used when "
                    "targets contain negative values."
                ):
                    continue
                else:
                    raise e

    else:
        for metric_ in to_score:

            # TODO maybe annotate metrics to define which cases they can
            # handle?

            try:
                score_dict[metric_.name] = _compute_single_scorer(
                    metric=metric_,
                    prediction=prediction,
                    solution=solution,
                    task_type=task_type,
                    X_data=X_data,
                )
            except ValueError as e:
                if e.args[0] == "multiclass format is not supported":
                    continue
                elif (
                    e.args[0] == "Samplewise metrics are not available "
                    "outside of multilabel classification."
                ):
                    continue
                elif (
                    e.args[0] == "Target is multiclass but "
                    "average='binary'. Please choose another average "
                    "setting, one of [None, 'micro', 'macro', 'weighted']."
                ):
                    continue
                else:
                    raise e

    return score_dict


def calculate_loss(
    solution: np.ndarray,
    prediction: np.ndarray,
    task_type: int,
    metric: Scorer,
    X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
) -> float:
    """Calculate the loss with a given metric

    Parameters
    ----------
    solution: np.ndarray
        The solutions

    prediction: np.ndarray
        The predictions generated

    task_type: int
        The task type of the problem

    metric: Scorer
        The metric to use

    X_data: Optional[SUPPORTED_XDATA_TYPES]
        X data used to obtain the predictions
    """
    losses = calculate_losses(
        solution=solution,
        prediction=prediction,
        task_type=task_type,
        metrics=[metric],
        X_data=X_data,
    )
    return losses[metric.name]


def calculate_losses(
    solution: np.ndarray,
    prediction: np.ndarray,
    task_type: int,
    metrics: Sequence[Scorer],
    *,
    X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
    scoring_functions: Optional[List[Scorer]] = None,
) -> Dict[str, float]:
    """
    Returns the losses (a magnitude that allows casting the
    optimization problem as a minimization one) for the
    given Auto-Sklearn Scorer objects.

    Parameters
    ----------
    solution: np.ndarray
        The ground truth of the targets
    prediction: np.ndarray
        The best estimate from the model, of the given targets
    task_type: int
        To understand if the problem task is classification
        or regression
    metrics: Sequence[Scorer]
        A list of objects that hosts a function to calculate how good the
        prediction is according to the solution.
    X_data: Optional[SUPPORTED_XDATA_TYPES]
        X data used to obtain the predictions
    scoring_functions: List[Scorer]
        A list of metrics to calculate multiple losses

    Returns
    -------
    Dict[str, float]
        A loss function for each of the provided scorer objects
    """
    score = calculate_scores(
        solution=solution,
        prediction=prediction,
        X_data=X_data,
        task_type=task_type,
        metrics=metrics,
        scoring_functions=scoring_functions,
    )
    scoring_functions = scoring_functions if scoring_functions else []

    # we expect a dict() object for which we should calculate the loss
    loss_dict = dict()
    for metric_ in scoring_functions + list(metrics):
        # maybe metric argument is not in scoring_functions
        # TODO: When metrics are annotated with type_of_target support
        # we can remove this check
        if metric_.name not in score:
            continue
        loss_dict[metric_.name] = metric_._optimum - score[metric_.name]
    return loss_dict


def compute_single_metric(
    metric: Scorer,
    prediction: np.ndarray,
    solution: np.ndarray,
    task_type: int,
    X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
) -> float:
    """
    Returns a metric for the given Auto-Sklearn Scorer object.
    It's direction is determined by the metric itself.

    Parameters
    ----------
    solution: np.ndarray
        The ground truth of the targets
    prediction: np.ndarray
        The best estimate from the model, of the given targets
    task_type: int
        To understand if the problem task is classification
        or regression
    metric: Scorer
        Object that host a function to calculate how good the
        prediction is according to the solution.
    X_data : array-like [n_samples x n_features]
        X data used to obtain the predictions

    Returns
    -------
    float
    """
    score = _compute_single_scorer(
        solution=solution,
        prediction=prediction,
        metric=metric,
        X_data=X_data,
        task_type=task_type,
    )
    return metric._sign * score


def _compute_single_scorer(
    metric: Scorer,
    prediction: np.ndarray,
    solution: np.ndarray,
    task_type: int,
    X_data: Optional[SUPPORTED_XDATA_TYPES] = None,
) -> float:
    """
    Returns a score (a magnitude that allows casting the
    optimization problem as a maximization one) for the
    given Auto-Sklearn Scorer object

    Parameters
    ----------
    solution: np.ndarray
        The ground truth of the targets
    prediction: np.ndarray
        The best estimate from the model, of the given targets
    task_type: int
        To understand if the problem task is classification
        or regression
    metric: Scorer
        Object that host a function to calculate how good the
        prediction is according to the solution.
    X_data : array-like [n_samples x n_features]
        X data used to obtain the predictions
    Returns
    -------
    float
    """
    if metric._needs_X:
        if X_data is None:
            raise ValueError(
                f"Metric {metric.name} needs X_data, but X_data is {X_data}"
            )
        elif X_data.shape[0] != solution.shape[0]:
            raise ValueError(
                f"X_data has wrong length. "
                f"Should be {solution.shape[0]}, but is {X_data.shape[0]}"
            )
        if task_type in REGRESSION_TASKS:
            # TODO put this into the regression metric itself
            cprediction = sanitize_array(prediction)
            score = metric(solution, cprediction, X_data=X_data)
        else:
            score = metric(solution, prediction, X_data=X_data)
        return score

    if task_type in REGRESSION_TASKS:
        # TODO put this into the regression metric itself
        cprediction = sanitize_array(prediction)
        score = metric(solution, cprediction)
    else:
        score = metric(solution, prediction)
    return score


# Must be at bottom so all metrics are defined
default_metric_for_task: Dict[int, Scorer] = {
    BINARY_CLASSIFICATION: CLASSIFICATION_METRICS["accuracy"],
    MULTICLASS_CLASSIFICATION: CLASSIFICATION_METRICS["accuracy"],
    MULTILABEL_CLASSIFICATION: CLASSIFICATION_METRICS["f1_macro"],
    REGRESSION: REGRESSION_METRICS["r2"],
    MULTIOUTPUT_REGRESSION: REGRESSION_METRICS["r2"],
}



# -*- encoding: utf-8 -*-

import os
import warnings

import numpy as np

__all__ = ["check_pid", "warn_if_not_float"]


def warn_if_not_float(X: np.ndarray, estimator: str = "This algorithm") -> bool:
    """Warning utility function to check that data type is floating point.
    Returns True if a warning was raised (i.e. the input is not float) and
    False otherwise, for easier input validation.
    """
    if not isinstance(estimator, str):
        estimator = estimator.__class__.__name__
    if X.dtype.kind != "f":
        warnings.warn(
            "%s assumes floating point values as input, "
            "got %s" % (estimator, X.dtype)
        )
        return True
    return False


def check_pid(pid: int) -> bool:
    """Check For the existence of a unix pid."""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def check_true(p: str) -> bool:
    if p in ("True", "true", 1, True):
        return True
    return False


def check_false(p: str) -> bool:
    if p in ("False", "false", 0, False):
        return True
    return False


def check_none(p: str) -> bool:
    if p in ("None", "none", None):
        return True
    return False


def check_for_bool(p: str) -> bool:
    if check_false(p):
        return False
    elif check_true(p):
        return True
    else:
        raise ValueError("%s is not a bool" % str(p))



import numpy as np
import pandas as pd
import sklearn.datasets
import sklearn.model_selection
from pandas.api.types import is_bool_dtype, is_numeric_dtype
from scipy import sparse
from sklearn.utils.multiclass import type_of_target

from autosklearn.data.target_validator import TargetValidator

import pytest


# Fixtures to be used in this class. By default all elements have 100 datapoints
@pytest.fixture
def input_data_targettest(request):
    if request.param == "series_binary":
        return pd.Series([1, -1, -1, 1])
    elif request.param == "series_multiclass":
        return pd.Series([1, 0, 2])
    elif request.param == "series_multilabel":
        return pd.Series([[1, 0], [0, 1]])
    elif request.param == "series_continuous":
        return pd.Series([0.1, 0.6, 0.7])
    elif request.param == "series_continuous-multioutput":
        return pd.Series([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == "pandas_binary":
        return pd.DataFrame([1, -1, -1, 1])
    elif request.param == "pandas_multiclass":
        return pd.DataFrame([1, 0, 2])
    elif request.param == "pandas_multilabel":
        return pd.DataFrame([[1, 0], [0, 1]])
    elif request.param == "pandas_continuous":
        return pd.DataFrame([0.1, 0.6, 0.7])
    elif request.param == "pandas_continuous-multioutput":
        return pd.DataFrame([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == "numpy_binary":
        return np.array([1, -1, -1, 1])
    elif request.param == "numpy_multiclass":
        return np.array([1, 0, 2])
    elif request.param == "numpy_multilabel":
        return np.array([[1, 0], [0, 1]])
    elif request.param == "numpy_continuous":
        return np.array([0.1, 0.6, 0.7])
    elif request.param == "numpy_continuous-multioutput":
        return np.array([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == "list_binary":
        return [1, -1, -1, 1]
    elif request.param == "list_multiclass":
        return [1, 0, 2]
    elif request.param == "list_multilabel":
        return [[0, 1], [1, 0]]
    elif request.param == "list_continuous":
        return [0.1, 0.6, 0.7]
    elif request.param == "list_continuous-multioutput":
        return [[1.5, 2.0], [3.0, 1.6]]
    elif "openml" in request.param:
        _, openml_id = request.param.split("_")
        X, y = sklearn.datasets.fetch_openml(
            data_id=int(openml_id), return_X_y=True, as_frame=True
        )
        if (
            len(y.shape) > 1
            and y.shape[1] > 1
            and np.any(y.eq("TRUE").any(1).to_numpy())
        ):
            # This 'if' is only asserted for multi-label data
            # Force the downloaded data to be interpreted as multilabel
            y = y.dropna()
            y.replace("FALSE", 0, inplace=True)
            y.replace("TRUE", 1, inplace=True)
            y = y.astype(int)
        return y
    elif "sparse" in request.param:
        # We expect the names to be of the type sparse_csc_nonan
        sparse_, type_, nan_ = request.param.split("_")
        if "nonan" in nan_:
            data = np.ones(3)
        else:
            data = np.array([1, 2, np.nan])

        # Then the type of sparse
        if "csc" in type_:
            return sparse.csc_matrix(data)
        elif "csr" in type_:
            return sparse.csr_matrix(data)
        elif "coo" in type_:
            return sparse.coo_matrix(data)
        elif "bsr" in type_:
            return sparse.bsr_matrix(data)
        elif "lil" in type_:
            return sparse.lil_matrix(data)
        elif "dok" in type_:
            return sparse.dok_matrix(np.vstack((data, data, data)))
        elif "dia" in type_:
            return sparse.dia_matrix(np.vstack((data, data, data)))
        else:
            ValueError("Unsupported indirect fixture {}".format(request.param))
    else:
        ValueError("Unsupported indirect fixture {}".format(request.param))


# Actual checks for the targets
@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_binary",
        "series_multiclass",
        "series_continuous",
        "pandas_binary",
        "pandas_multiclass",
        "pandas_multilabel",
        "pandas_continuous",
        "pandas_continuous-multioutput",
        "numpy_binary",
        "numpy_multiclass",
        "numpy_multilabel",
        "numpy_continuous",
        "numpy_continuous-multioutput",
        "list_binary",
        "list_multiclass",
        "list_multilabel",
        "list_continuous",
        "list_continuous-multioutput",
        "openml_204",
    ),
    indirect=True,
)
def test_targetvalidator_supported_types_noclassification(input_data_targettest):
    y = input_data_targettest

    validator = TargetValidator(is_classification=False)
    validator.fit(input_data_targettest)

    y_encoded: np.ndarray = validator.transform(y)

    # Assert the validator was fitted
    assert validator._is_fitted

    # Assert no encoder if is_classification = False
    assert validator.encoder is None

    # Assert it returns y_inverse, y_encoded as ndarray
    assert isinstance(y_encoded, np.ndarray)

    # Assert the return value is equal to the one passed in
    np.testing.assert_array_equal(y_encoded, y)


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_binary",
        "series_multiclass",
        "pandas_binary",
        "pandas_multiclass",
        "numpy_binary",
        "numpy_multiclass",
        "list_binary",
        "list_multiclass",
        "openml_2",
    ),
    indirect=True,
)
def test_targetvalidator_supported_types_classification(input_data_targettest):
    y = input_data_targettest  # Just to remove visual clutter

    validator = TargetValidator(is_classification=True)
    validator.fit(y)

    y_encoded: np.ndarray = validator.transform(y)
    y_inverse: np.ndarray = validator.inverse_transform(y_encoded)

    # Assert the validator was fitted
    assert validator._is_fitted

    # Assert it returns y_inverse, y_encoded as ndarray
    assert isinstance(y_encoded, np.ndarray)
    assert isinstance(y_inverse, np.ndarray)

    # Assert that y_encoded is numeric and not boolean
    assert is_numeric_dtype(y_encoded.dtype) and not is_bool_dtype(y_encoded.dtype)

    # Assert dtype is presevered with y -> y_encoded -> y_inverse
    def dtype(arr):
        if isinstance(arr, list):
            if isinstance(arr[0], list):
                return type(arr[0][0])
            else:
                return type(arr[0])
        elif isinstance(arr, pd.DataFrame):
            return arr.dtypes[0]
        else:
            return arr.dtype

    if is_numeric_dtype(dtype(y)):
        assert y_inverse.dtype == dtype(y)

    # Assert that y == y_inverse, giving back the original values
    np.testing.assert_array_equal(y_inverse, np.array(y).reshape(-1))

    # Assert the y_inverse, y_encoded have the expected shape
    shape = np.shape(y)
    if len(shape) == 2 and shape[1] == 1:
        # For cases where y = [[1], [2], [3]],
        # we expect y_inverse, y_encodedd to have been flattened to [1,2,3]
        expected_shape = (shape[0],)
    else:
        expected_shape = shape

    assert y_encoded.shape == expected_shape
    assert y_inverse.shape == expected_shape

    # These next part of the tests rely on some encoding to have taken place
    # This happens when `is_classification` and not task_type = multilabel-indicator
    #
    # As state in TargetValidator._fit()
    # > Also, encoding multilabel indicator data makes the data multiclass
    #   Let the user employ a MultiLabelBinarizer if needed
    #
    # As a result of this, we don't encode 'multilabel-indicator' labels and
    # there is nothing else to check here
    if validator.type_of_target == "multilabel-indicator":
        assert validator.encoder is None

    else:
        assert validator.encoder is not None

        # Assert the columns have had their labels converted
        if y_encoded.ndim == 1:
            max_label = len(np.unique(y_encoded)) - 1
            assert np.min(y_encoded) == 0
            assert np.max(y_encoded) == max_label

        else:
            for col in range(y_encoded.shape[1]):
                column = y_encoded[:, col]
                max_label = len(np.unique(y_encoded[:, col])) - 1
                assert np.min(column) == 0
                assert np.max(column) == max_label


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_binary",
        "pandas_binary",
        "numpy_binary",
        "list_binary",
        "openml_1066",
    ),
    indirect=True,
)
def test_targetvalidator_binary(input_data_targettest):
    assert type_of_target(input_data_targettest) == "binary"
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == "binary"


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_multiclass",
        "pandas_multiclass",
        "numpy_multiclass",
        "list_multiclass",
        "openml_54",
    ),
    indirect=True,
)
def test_targetvalidator_multiclass(input_data_targettest):
    assert type_of_target(input_data_targettest) == "multiclass"
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == "multiclass"


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "pandas_multilabel",
        "numpy_multilabel",
        "list_multilabel",
        "openml_40594",
    ),
    indirect=True,
)
def test_targetvalidator_multilabel(input_data_targettest):
    assert type_of_target(input_data_targettest) == "multilabel-indicator"
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == "multilabel-indicator"


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_continuous",
        "pandas_continuous",
        "numpy_continuous",
        "list_continuous",
        "openml_531",
    ),
    indirect=True,
)
def test_targetvalidator_continuous(input_data_targettest):
    assert type_of_target(input_data_targettest) == "continuous"
    validator = TargetValidator(is_classification=False)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == "continuous"


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "pandas_continuous-multioutput",
        "numpy_continuous-multioutput",
        "list_continuous-multioutput",
        "openml_41483",
    ),
    indirect=True,
)
def test_targetvalidator_continuous_multioutput(input_data_targettest):
    assert type_of_target(input_data_targettest) == "continuous-multioutput"
    validator = TargetValidator(is_classification=False)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == "continuous-multioutput"


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_binary",
        "pandas_binary",
        "numpy_binary",
        "list_binary",
    ),
    indirect=True,
)
def test_targetvalidator_fitontypeA_transformtypeB(input_data_targettest):
    """
    Check if we can fit in a given type (numpy) yet transform
    if the user changes the type (pandas then)

    This is problematic only in the case we create an encoder
    """
    validator = TargetValidator(is_classification=True)
    validator.fit(input_data_targettest)
    if isinstance(input_data_targettest, pd.DataFrame):
        complementary_type = input_data_targettest.to_numpy()
    elif isinstance(input_data_targettest, pd.Series):
        complementary_type = pd.DataFrame(input_data_targettest)
    elif isinstance(input_data_targettest, np.ndarray):
        complementary_type = pd.DataFrame(input_data_targettest)
    elif isinstance(input_data_targettest, list):
        complementary_type = pd.DataFrame(input_data_targettest)
    validator.transform(complementary_type)


@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_multilabel",
        "series_continuous-multioutput",
    ),
    indirect=True,
)
def test_type_of_target_unsupported(input_data_targettest):
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input
    """
    validator = TargetValidator()
    with pytest.raises(ValueError, match=r"legacy multi-.* data representation."):
        validator.fit(input_data_targettest)


def test_target_unsupported():
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input
    """
    validator = TargetValidator(is_classification=True)
    with pytest.raises(
        ValueError, match=r"The dimensionality of the train and test targets"
    ):
        validator.fit(
            np.array([[0, 1, 0], [0, 1, 1]]),
            np.array([[0, 1, 0, 0], [0, 1, 1, 1]]),
        )
    with pytest.raises(
        ValueError, match=r"Train and test targets must both have the same dtypes"
    ):
        validator.fit(
            pd.DataFrame({"a": [1, 2, 3]}),
            pd.DataFrame({"a": [True, False, False]}),
        )
    with pytest.raises(ValueError, match=r"Provided targets are not supported.*"):
        validator.fit(
            np.array([[0, 1, 2], [0, 3, 4]]),
            np.array([[0, 1, 2, 5], [0, 3, 4, 6]]),
        )
    with pytest.raises(
        ValueError, match="Train and test targets must both have the same"
    ):
        validator.fit(
            pd.DataFrame({"string": ["foo"]}),
            pd.DataFrame({"int": [1]}),
        )
    with pytest.raises(
        ValueError, match=r"Auto-sklearn only supports Numpy arrays, .*"
    ):
        validator.fit({"input1": 1, "input2": 2})
    with pytest.raises(
        ValueError, match=r"arget values cannot contain missing/NaN values"
    ):
        validator.fit(np.array([np.nan, 1, 2]))
    with pytest.raises(
        ValueError, match=r"arget values cannot contain missing/NaN values"
    ):
        validator.fit(sparse.csr_matrix(np.array([1, 2, np.nan])))
    with pytest.raises(
        ValueError, match=r"TargetValidator must have fit\(\) called first"
    ):
        validator.transform(np.array([1, 2, 3]))
    with pytest.raises(
        ValueError, match=r"TargetValidator must have fit\(\) called first"
    ):
        validator.inverse_transform(np.array([1, 2, 3]))
    with pytest.raises(
        ValueError, match=r"Multi-dimensional classification is not yet supported"
    ):
        validator._fit(np.array([[1, 2, 3], [1, 5, 6]]))

    # Dia/ DOK are not supported as type of target makes calls len on the array
    # which causes TypeError: len() of unsized object. Basically, sparse data as
    # multi-label is the only thing that makes sense in this format.
    with pytest.raises(
        ValueError, match=r"The provided data could not be interpreted by Sklearn"
    ):
        validator.fit(sparse.dia_matrix(np.array([1, 2, 3])))

    validator.fit(np.array([[0, 1, 0], [0, 1, 1]]))
    with pytest.raises(ValueError, match=r"Number of outputs changed from"):
        validator.fit(np.array([0, 1, 0]))


def test_targetvalidator_inversetransform():
    """
    Test that the encoding/decoding works in 1D
    """
    validator = TargetValidator(is_classification=True)
    validator.fit(
        pd.DataFrame(data=["a", "a", "b", "c", "a"], dtype="category"),
    )
    y = validator.transform(
        pd.DataFrame(data=["a", "a", "b", "c", "a"], dtype="category"),
    )
    np.testing.assert_array_almost_equal(np.array([0, 0, 1, 2, 0]), y)

    y_decoded = validator.inverse_transform(y)
    assert ["a", "a", "b", "c", "a"] == y_decoded.tolist()

    assert validator.classes_.tolist() == ["a", "b", "c"]

    validator = TargetValidator(is_classification=True)
    multi_label = pd.DataFrame(
        np.array([[1, 0, 0, 1], [0, 0, 1, 1], [0, 0, 0, 0]]), dtype=bool
    )
    validator.fit(multi_label)
    y = validator.transform(multi_label)

    y_decoded = validator.inverse_transform(y)
    np.testing.assert_array_almost_equal(y, y_decoded)

    # Multilabel classification is not encoded
    # For this reason, classes_ attribute does not contain a class
    np.testing.assert_array_almost_equal(validator.classes_, np.array([]))


# Actual checks for the targets
@pytest.mark.parametrize(
    "input_data_targettest",
    (
        "series_binary",
        "series_multiclass",
        "pandas_binary",
        "pandas_multiclass",
        "numpy_binary",
        "numpy_multiclass",
        "list_binary",
        "list_multiclass",
    ),
    indirect=True,
)
def test_unknown_categories_in_targets(input_data_targettest):
    validator = TargetValidator(is_classification=True)
    validator.fit(input_data_targettest)

    # Add an extra category
    if isinstance(input_data_targettest, list):
        input_data_targettest.append(input_data_targettest[-1] + 5000)
    elif isinstance(input_data_targettest, (pd.DataFrame, pd.Series)):
        input_data_targettest.iloc[-1] = 5000
    elif isinstance(input_data_targettest, np.ndarray):
        input_data_targettest[-1] = 5000

    with pytest.raises(ValueError, match="Found unknown categories"):
        validator.transform(input_data_targettest)


def test_is_single_column_target():
    validator = TargetValidator(is_classification=True)
    validator.fit(np.array([1, 2, 3, 4]))
    assert validator.is_single_column_target()

    validator = TargetValidator(is_classification=True)
    validator.fit(np.array([[1, 0, 1, 0], [1, 1, 1, 1]]))
    assert not validator.is_single_column_target()



from typing import Optional

from functools import partial

from ConfigSpace import NotEqualsCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class SelectClassificationRates(AutoSklearnPreprocessingAlgorithm):
    def __init__(self, alpha, mode="fpr", score_func="chi2", random_state=None):
        import sklearn.feature_selection

        self.random_state = random_state  # We don't use this
        self.alpha = alpha
        self.mode = mode

        if score_func == "chi2":
            self.score_func = sklearn.feature_selection.chi2
        elif score_func == "f_classif":
            self.score_func = sklearn.feature_selection.f_classif
        elif score_func == "mutual_info_classif":
            self.score_func = partial(
                sklearn.feature_selection.mutual_info_classif,
                random_state=self.random_state,
            )
            # mutual info classif constantly crashes without mode percentile
            self.mode = "percentile"
        else:
            raise ValueError(
                "score_func must be in ('chi2, 'f_classif', 'mutual_info_classif') "
                "for classification "
                "but is: %s " % (score_func)
            )

    def fit(self, X, y):
        import scipy.sparse
        import sklearn.feature_selection

        self.alpha = float(self.alpha)

        self.preprocessor = sklearn.feature_selection.GenericUnivariateSelect(
            score_func=self.score_func, param=self.alpha, mode=self.mode
        )

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.score_func == sklearn.feature_selection.chi2:
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        self.preprocessor.fit(X, y)
        return self

    def transform(self, X):
        import scipy.sparse
        import sklearn.feature_selection

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.score_func == sklearn.feature_selection.chi2:
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        if self.preprocessor is None:
            raise NotImplementedError()
        try:
            Xt = self.preprocessor.transform(X)
        except ValueError as e:
            if (
                "zero-size array to reduction operation maximum which has no "
                "identity" in e.message
            ):
                raise ValueError("%s removed all features." % self.__class__.__name__)
            else:
                raise e

        if Xt.shape[1] == 0:
            raise ValueError("%s removed all features." % self.__class__.__name__)
        return Xt

    @staticmethod
    def get_properties(dataset_properties=None):
        data_type = UNSIGNED_DATA

        if dataset_properties is not None:
            signed = dataset_properties.get("signed")
            if signed is not None:
                data_type = SIGNED_DATA if signed is True else UNSIGNED_DATA

        return {
            "shortname": "SR",
            "name": "Univariate Feature Selection based on rates",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (SPARSE, DENSE, data_type),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        alpha = UniformFloatHyperparameter(
            name="alpha", lower=0.01, upper=0.5, default_value=0.1
        )

        if dataset_properties is not None and dataset_properties.get("sparse"):
            choices = ["chi2", "mutual_info_classif"]
        else:
            choices = ["chi2", "f_classif", "mutual_info_classif"]

        score_func = CategoricalHyperparameter(
            name="score_func", choices=choices, default_value="chi2"
        )

        mode = CategoricalHyperparameter("mode", ["fpr", "fdr", "fwe"], "fpr")

        cs = ConfigurationSpace()
        cs.add_hyperparameter(alpha)
        cs.add_hyperparameter(score_func)
        cs.add_hyperparameter(mode)

        # mutual_info_classif constantly crashes if mode is not percentile
        # as a WA, fix the mode for this score
        cond = NotEqualsCondition(mode, score_func, "mutual_info_classif")
        cs.add_condition(cond)

        return cs



import sklearn.naive_bayes

from autosklearn.pipeline.components.classification.gaussian_nb import GaussianNB

from .test_base import BaseClassificationComponentTest


class GaussianNBComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.95999999999999996
    res["iris_n_calls"] = None
    res["default_iris_iterative"] = 0.95999999999999996
    res["default_iris_proba"] = 0.11199001987342033
    res["default_iris_sparse"] = -1
    res["default_digits"] = 0.80692167577413476
    res["digits_n_calls"] = None
    res["default_digits_iterative"] = 0.80692167577413476
    res["default_digits_binary"] = 0.98664238008500305
    res["default_digits_multilabel"] = 0.54135471896765841
    res["default_digits_multilabel_proba"] = 0.99028976450984096

    sk_mod = sklearn.naive_bayes.GaussianNB
    module = GaussianNB



# -*- encoding: utf-8 -*-
import os

from autosklearn.util.common import check_pid

import unittest


class TestUtilsCommon(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_check_pid(self):
        our_pid = os.getpid()

        exists = check_pid(our_pid)
        self.assertTrue(exists)
        our_pid = -11000  # We hope this pid does not exist
        exists = check_pid(our_pid)
        self.assertFalse(exists)


if __name__ == "__main__":
    unittest.main()



"""Test the show models functinality of an automl instance"""
from autosklearn.automl import AutoML

from pytest_cases import parametrize_with_cases

import test.test_automl.cases as cases


@parametrize_with_cases("automl", cases=cases, has_tag=["fitted", "no_ensemble"])
def test_no_ensemble_produces_empty_show_models(automl: AutoML) -> None:
    """
    Parameters
    ----------
    automl : AutoML
        The automl object with no ensemble size to test

    Expects
    -------
    * Show models should return an empty dict
    """
    assert automl.show_models() == {}



import ast

import numpy as np
import pandas as pd
import sklearn.utils

from autosklearn.metalearning.metalearning.kNearestDatasets.kND import KNearestDatasets


def test_function(params):
    return np.random.random()


class MetaLearningOptimizer(object):
    def __init__(
        self,
        dataset_name,
        configuration_space,
        meta_base,
        logger,
        distance="l1",
        seed=None,
        use_features=None,
        distance_kwargs=None,
    ):
        self.dataset_name = dataset_name
        self.configuration_space = configuration_space
        self.meta_base = meta_base
        self.distance = distance
        self.seed = seed
        self.use_features = use_features
        self.distance_kwargs = distance_kwargs
        self.kND = None  # For caching, makes things faster...

        self.logger = logger

    def metalearning_suggest_all(self, exclude_double_configurations=True):
        """Return a list of the best hyperparameters of neighboring datasets"""
        # TODO check if _learn was called before!
        neighbors = self._learn(exclude_double_configurations)
        hp_list = []
        for neighbor in neighbors:
            try:
                configuration = self.meta_base.get_configuration_from_algorithm_index(
                    neighbor[2]
                )
                self.logger.info("%s %s %s" % (neighbor[0], neighbor[1], configuration))
            except (KeyError):
                self.logger.warning("Configuration %s not found" % neighbor[2])
                continue

            hp_list.append(configuration)
        return hp_list

    def metalearning_suggest(self, history):
        """Suggest the next promosing hyperparameters which were not yet evaluated"""
        # TODO test the object in the history!
        neighbors = self._learn()
        # Iterate over all datasets which are sorted ascending by distance

        history_with_indices = []
        for run in history:
            history_with_indices.append(
                self.meta_base.get_algorithm_index_from_configuration(run)
            )

        for idx, neighbor in enumerate(neighbors):
            already_evaluated = False
            # Check if that dataset was already evaluated
            for run in history_with_indices:
                # If so, return to the outer loop

                if neighbor[2] == run:
                    already_evaluated = True
                    break

            if not already_evaluated:
                self.logger.info(
                    "Nearest dataset with hyperparameters of best value "
                    "not evaluated yet is %s with a distance of %f"
                    % (neighbor[0], neighbor[1])
                )
                return self.meta_base.get_configuration_from_algorithm_index(
                    neighbor[2]
                )
        raise StopIteration("No more values available.")

    def _learn(self, exclude_double_configurations=True):
        dataset_metafeatures, all_other_metafeatures = self._split_metafeature_array()

        # Remove metafeatures which could not be calculated for the target
        # dataset
        keep = []
        for idx in dataset_metafeatures.index:
            if np.isfinite(dataset_metafeatures.loc[idx]):
                keep.append(idx)

        dataset_metafeatures = dataset_metafeatures.loc[keep]
        all_other_metafeatures = all_other_metafeatures.loc[:, keep]

        # Do mean imputation of all other metafeatures
        all_other_metafeatures = all_other_metafeatures.fillna(
            all_other_metafeatures.mean()
        )

        if self.kND is None:
            # In case that we learn our distance function, get_value the parameters for
            #  the random forest
            if self.distance_kwargs:
                rf_params = ast.literal_eval(self.distance_kwargs)
            else:
                rf_params = None

            # To keep the distance the same in every iteration, we create a new
            # random state
            random_state = sklearn.utils.check_random_state(self.seed)
            kND = KNearestDatasets(
                metric=self.distance,
                random_state=random_state,
                logger=self.logger,
                metric_params=rf_params,
            )

            runs = dict()
            # TODO move this code to the metabase
            for task_id in all_other_metafeatures.index:
                try:
                    runs[task_id] = self.meta_base.get_runs(task_id)
                except KeyError:
                    # TODO should I really except this?
                    self.logger.info("Could not find runs for instance %s" % task_id)
                    runs[task_id] = pd.Series([], name=task_id, dtype=np.float64)

            runs = pd.DataFrame(runs)

            kND.fit(all_other_metafeatures, runs)
            self.kND = kND
        return self.kND.kBestSuggestions(
            dataset_metafeatures,
            k=-1,
            exclude_double_configurations=exclude_double_configurations,
        )

    def _split_metafeature_array(self):
        dataset_metafeatures = self.meta_base.get_metafeatures(
            self.dataset_name, self.use_features
        )
        all_other_datasets = self.meta_base.get_all_dataset_names()
        all_other_datasets.remove(self.dataset_name)
        all_other_metafeatures = self.meta_base.get_metafeatures(
            all_other_datasets, self.use_features
        )
        return dataset_metafeatures, all_other_metafeatures



"""Test the dummy predictor of AutoML

Dummy models can serve as an early warning of issues with parameters during fit
"""
from __future__ import annotations

from typing import Callable, Sequence, Tuple

from pathlib import Path

import numpy as np
from smac.tae import StatusType

from autosklearn.automl import AutoML
from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    REGRESSION,
)
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.metrics import Scorer, accuracy, log_loss, precision, r2
from autosklearn.util.logging_ import PicklableClientLogger

import pytest
from pytest_cases import parametrize
from unittest.mock import patch


@parametrize(
    "dataset, metrics, task",
    [
        ("breast_cancer", [accuracy], BINARY_CLASSIFICATION),
        ("breast_cancer", [accuracy, log_loss], BINARY_CLASSIFICATION),
        ("wine", [accuracy], MULTICLASS_CLASSIFICATION),
        ("diabetes", [r2], REGRESSION),
    ],
)
def test_produces_correct_output(
    dataset: str,
    task: int,
    metrics: Sequence[Scorer],
    mock_logger: PicklableClientLogger,
    make_automl: Callable[..., AutoML],
    make_sklearn_dataset: Callable[..., XYDataManager],
) -> None:
    """
    Parameters
    ----------
    dataset: str
        The name of the dataset

    task : int
        The task type of the dataset

    metrics: Sequence[Scorer]
        Metric(s) to use, required as fit usually determines the metric to use

    Fixtures
    --------
    mock_logger: PickleableClientLogger
        A mock logger to use

    make_automl : Callable[..., AutoML]
        Factory to make an AutoML object

    make_sklearn_dataset : Callable[..., XYDataManager]
        Factory to get an sklearn dataset

    Expects
    -------
    * There should only be one output created with one dummy predictions
    * It should be named "1337_1_0.0" with {seed}_{num_run}_{budget}
    * It should produce predictions "predictions_ensemble_1337_1_0.0.npy"
    """
    seed = 1337
    automl = make_automl(metrics=metrics, seed=seed)
    automl._logger = mock_logger

    datamanager = make_sklearn_dataset(
        dataset,
        as_datamanager=True,
        task=task,
        feat_type="numerical",
    )
    automl._backend.save_datamanager(datamanager)
    automl._do_dummy_prediction()

    path = Path(automl._backend.get_runs_directory())
    run_paths = list(path.iterdir())
    assert len(run_paths) == 1

    dummy_run_path = run_paths[0]
    assert dummy_run_path.name == f"{seed}_1_0.0"

    predictions_path = dummy_run_path / f"predictions_ensemble_{seed}_1_0.0.npy"
    assert predictions_path.exists()


def test_runs_with_correct_args(
    mock_logger: PicklableClientLogger,
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
    make_automl: Callable[..., AutoML],
) -> None:
    """
    Fixtures
    --------
    mock_logger: PickleableClientLogger
        A mock logger to use

    make_sklearn_dataset : Callable[..., Tuple[np.ndarray, ...]]
        Factory to make dataset

    make_automl : Callable[..., AutoML]
        Factory to make automl

    Expects
    -------
    * The mock run should be called once with:
        * config = 1 (The always given number for the dummy)
        * cutoff = `automl._time_for_task` (the fulll time for the task)
    """
    dataset = "iris"
    task = MULTICLASS_CLASSIFICATION

    automl = make_automl(metrics=[accuracy])
    automl._logger = mock_logger

    datamanager = make_sklearn_dataset(
        dataset,
        as_datamanager=True,
        task=task,
        feat_type="numerical",
    )
    automl._backend.save_datamanager(datamanager)

    with patch("autosklearn.evaluation.ExecuteTaFuncWithQueue.run") as mck:
        mck.return_value = (StatusType.SUCCESS, None, None, {})
        automl._do_dummy_prediction()

    mck.assert_called_once_with(config=1, cutoff=automl._time_for_task)


def test_crash_due_to_memory_exception(
    mock_logger: PicklableClientLogger,
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
    make_automl: Callable[..., AutoML],
) -> None:
    """
    Fixtures
    --------
    mock_logger: PickleableClientLogger
        A mock logger to use

    make_sklearn_dataset : Callable[..., Tuple[np.ndarray, ...]]
        Factory to make dataset

    make_automl : Callable[..., AutoML]
        Factory to make automl

    Expects
    -------
    * The dummy prediction should raise when encoutering with StatusType.CRASHED
    * The error message should indicate it's a memory issue with `{'exitcode' -6}`
      encountered
    """
    dataset = "iris"
    task = MULTICLASS_CLASSIFICATION

    automl = make_automl(metrics=[accuracy])
    automl._logger = mock_logger

    datamanager = make_sklearn_dataset(
        dataset,
        as_datamanager=True,
        task=task,
        feat_type="numerical",
    )

    automl._backend.save_datamanager(datamanager)

    with patch("autosklearn.evaluation.ExecuteTaFuncWithQueue.run") as mck:
        mck.return_value = (StatusType.CRASHED, None, None, {"exitcode": -6})
        msg = "The error suggests that the provided memory limits are too tight."

        with pytest.raises(ValueError, match=msg):
            automl._do_dummy_prediction()


def test_raises_if_no_metric_set(make_automl: Callable[..., AutoML]) -> None:
    """
    Expects
    -------
    * raise if there was no metric set when calling `_do_dummy_prediction()`
    """
    automl = make_automl()
    with pytest.raises(ValueError, match="Metric/Metrics was/were not set"):
        automl._do_dummy_prediction()


def test_raises_invalid_metric(
    mock_logger: PicklableClientLogger,
    make_automl: Callable[..., AutoML],
    make_sklearn_dataset: Callable[..., XYDataManager],
) -> None:
    """
    Expects
    -------
    * Should raise an error if the given metric is not applicable to a given task type
    """
    # `precision` is not applicable to MULTICLASS_CLASSIFICATION
    dataset = "iris"
    task = MULTICLASS_CLASSIFICATION
    metrics = [accuracy, precision]

    automl = make_automl(metrics=metrics)
    automl._logger = mock_logger

    datamanager = make_sklearn_dataset(
        dataset,
        as_datamanager=True,
        task=task,
        feat_type="numerical",
    )
    automl._backend.save_datamanager(datamanager)

    with pytest.raises(
        ValueError,
        match="Are you sure precision is applicable for the given task type",
    ):
        automl._do_dummy_prediction()



import os
import tempfile

import pandas as pd
from smac.callbacks import IncorporateRunResultCallback
from smac.optimizer.smbo import SMBO
from smac.runhistory.runhistory import RunInfo, RunValue
from smac.tae.base import StatusType

import autosklearn.pipeline.util as putil
from autosklearn.classification import AutoSklearnClassifier

import unittest


class AutoMLTrialsCallBack(IncorporateRunResultCallback):
    def __init__(self, fname):
        self.trials_num = 1
        self.fname = fname
        with open(fname, "w") as fp:
            fp.write(
                "TrialNo, "
                "StartTime, "
                "EndTime, "
                "Status, "
                "TrainLoss, "
                "ValidLoss, "
                "TestLoss, "
                "Classifier"
            )

    def __call__(
        self,
        smbo: "SMBO",
        run_info: RunInfo,
        result: RunValue,
        time_left: float,
    ) -> None:
        train_loss, valid_loss, test_loss = None, None, None
        trial_start_time = result.starttime
        trial_end_time = result.endtime
        trial_status = result.status.name
        if trial_status == StatusType.SUCCESS.name:
            train_loss = result.additional_info.get("train_loss")
            valid_loss = result.cost
            test_loss = result.additional_info.get("test_loss")
        trial_classifier = run_info.config.get_dictionary()["classifier:__choice__"]
        with open(self.fname, "a+") as fp:
            fp.write(
                f"\n {self.trials_num}, {trial_start_time}, {trial_end_time},"
                f" {trial_status}, {train_loss}, {valid_loss}, {test_loss},"
                f" {trial_classifier}"
            )
        self.trials_num += 1


class VerifyTrialsCallBack(unittest.TestCase):
    def test_trials_callback_execution(self):
        trials_summary_fname = os.path.join(tempfile.gettempdir(), "trials.csv")
        X_train, Y_train, X_test, Y_test = putil.get_dataset("breast_cancer")
        cls = AutoSklearnClassifier(
            time_left_for_this_task=30,
            initial_configurations_via_metalearning=0,
            per_run_time_limit=10,
            delete_tmp_folder_after_terminate=False,
            n_jobs=1,
            include={"feature_preprocessor": ["pca"], "classifier": ["sgd"]},
            get_trials_callback=AutoMLTrialsCallBack(trials_summary_fname),
        )
        cls.fit(X_train, Y_train, X_test, Y_test)
        trials = pd.read_csv(trials_summary_fname)
        assert (
            trials.shape[0] > 0
        ), f"Auto-Sklearn explored {trials.shape[0] - 1} trials"



from __future__ import annotations

from typing import Callable

from autosklearn.automl import AutoML
from autosklearn.ensemble_building import EnsembleBuilderManager

from pytest_cases import parametrize_with_cases

import test.test_automl.cases as cases
from test.conftest import DEFAULT_SEED


@parametrize_with_cases("automl", cases=cases, has_tag="fitted")
def case_real_runs(
    automl: AutoML,
    make_ensemble_builder_manager: Callable[..., EnsembleBuilderManager],
) -> EnsembleBuilderManager:
    """Uses real runs from a fitted automl instance"""
    manager = make_ensemble_builder_manager(
        backend=automl._backend,
        metric=automl._metrics[0],
        task=automl._task,
        dataset_name=automl._dataset_name,
        seed=automl._seed,
        logger_port=automl._logger_port,
        random_state=DEFAULT_SEED,
    )
    return manager



import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator, TransformerMixin


class MinorityCoalescer(BaseEstimator, TransformerMixin):
    """Group together categories which occurence is less than a specified
    minimum fraction. Coalesced categories get index of one.
    """

    def __init__(self, minimum_fraction=None):
        self.minimum_fraction = minimum_fraction

    def check_X(self, X):
        X_data = X.data if sparse.issparse(X) else X
        if np.nanmin(X_data) < 2:
            raise ValueError("X needs to contain only integers greater than two.")

    def fit(self, X, y=None):
        self.check_X(X)

        if self.minimum_fraction is None:
            return self

        # Remember which values should not be coalesced
        do_not_coalesce = list()
        for column in range(X.shape[1]):
            do_not_coalesce.append(set())

            if sparse.issparse(X):
                indptr_start = X.indptr[column]
                indptr_end = X.indptr[column + 1]
                unique, counts = np.unique(
                    X.data[indptr_start:indptr_end], return_counts=True
                )
                colsize = indptr_end - indptr_start
            else:
                unique, counts = np.unique(X[:, column], return_counts=True)
                colsize = X.shape[0]

            for unique_value, count in zip(unique, counts):
                fraction = float(count) / colsize
                if fraction >= self.minimum_fraction:
                    do_not_coalesce[-1].add(unique_value)

        self.do_not_coalesce_ = do_not_coalesce
        return self

    def transform(self, X):
        self.check_X(X)

        if self.minimum_fraction is None:
            return X

        for column in range(X.shape[1]):
            if sparse.issparse(X):
                indptr_start = X.indptr[column]
                indptr_end = X.indptr[column + 1]
                unique = np.unique(X.data[indptr_start:indptr_end])
                for unique_value in unique:
                    if unique_value not in self.do_not_coalesce_[column]:
                        indptr_start = X.indptr[column]
                        indptr_end = X.indptr[column + 1]
                        X.data[indptr_start:indptr_end][
                            X.data[indptr_start:indptr_end] == unique_value
                        ] = 1
            else:
                unique = np.unique(X[:, column])
                unique_values = [
                    unique_value
                    for unique_value in unique
                    if unique_value not in self.do_not_coalesce_[column]
                ]
                mask = np.isin(X[:, column], unique_values)
                X[mask, column] = 1
        return X

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)



import sklearn.linear_model

from autosklearn.pipeline.components.regression.sgd import SGD

from .test_base import BaseRegressionComponentTest


class SGDComponentTest(BaseRegressionComponentTest):
    __test__ = True

    # Values are extremely bad because the invscaling does not drop the
    # learning rate aggressively enough!
    res = dict()
    res["default_boston"] = -1.1811672998629865e28
    res["boston_n_calls"] = 6
    res["default_boston_iterative"] = res["default_boston"]
    res["default_boston_sparse"] = -1.1518512489347601e28
    res["default_boston_iterative_sparse"] = res["default_boston_sparse"]
    res["default_diabetes"] = 0.27420813549185374
    res["diabetes_n_calls"] = 10
    res["default_diabetes_iterative"] = res["default_diabetes"]
    res["default_diabetes_sparse"] = 0.034801785011824404
    res["default_diabetes_iterative_sparse"] = res["default_diabetes_sparse"]

    sk_mod = sklearn.linear_model.SGDRegressor
    module = SGD



import logging
import os

import arff
import numpy as np
from scipy import sparse
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler

import autosklearn.metalearning.metafeatures.metafeatures as meta_features
from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)

import pytest


@pytest.fixture
def sparse_data():
    tests_dir = __file__
    os.chdir(os.path.dirname(tests_dir))

    decoder = arff.ArffDecoder()
    with open(os.path.join("datasets", "dataset.arff")) as fh:
        dataset = decoder.decode(fh, encode_nominal=True)

    # -1 because the last attribute is the class
    attribute_types = [
        "numeric" if type(type_) != list else "nominal"
        for name, type_ in dataset["attributes"][:-1]
    ]
    feat_type = {
        i: "categorical" if attribute == "nominal" else "numerical"
        for i, attribute in enumerate(attribute_types)
    }

    data = np.array(dataset["data"], dtype=np.float64)
    X = data[:, :-1]
    y = data[:, -1].reshape((-1,))

    # First, swap NaNs and zeros, because when converting an encoded
    # dense matrix to sparse, the values which are encoded to zero are lost
    X_sparse = X.copy()
    NaNs = ~np.isfinite(X_sparse)
    X_sparse[NaNs] = 0
    X_sparse = sparse.csr_matrix(X_sparse)

    X = X_sparse
    y = y
    mf = meta_features.metafeatures
    helpers = meta_features.helper_functions
    logger = logging.getLogger()
    # Precompute some helper functions
    helpers.set_value(
        "MissingValues",
        helpers["MissingValues"](X, y, logger, feat_type),
    )
    mf.set_value(
        "NumberOfMissingValues",
        mf["NumberOfMissingValues"](X, y, logger, feat_type),
    )
    helpers.set_value(
        "NumSymbols",
        helpers["NumSymbols"](X, y, logger, feat_type),
    )
    helpers.set_value(
        "ClassOccurences",
        helpers["ClassOccurences"](X, y, logger),
    )
    return X, y, feat_type


@pytest.fixture
def sparse_data_transformed():
    tests_dir = __file__
    os.chdir(os.path.dirname(tests_dir))

    decoder = arff.ArffDecoder()
    with open(os.path.join("datasets", "dataset.arff")) as fh:
        dataset = decoder.decode(fh, encode_nominal=True)

    # -1 because the last attribute is the class
    attribute_types = [
        "numeric" if type(type_) != list else "nominal"
        for name, type_ in dataset["attributes"][:-1]
    ]
    feat_type = {
        i: "categorical" if attribute == "nominal" else "numerical"
        for i, attribute in enumerate(attribute_types)
    }

    data = np.array(dataset["data"], dtype=np.float64)
    X = data[:, :-1]
    y = data[:, -1].reshape((-1,))

    # First, swap NaNs and zeros, because when converting an encoded
    # dense matrix to sparse, the values which are encoded to zero are lost
    X_sparse = X.copy()
    NaNs = ~np.isfinite(X_sparse)
    X_sparse[NaNs] = 0
    X_sparse = sparse.csr_matrix(X_sparse)

    ohe = FeatTypeSplit(feat_type=feat_type)
    X_transformed = X_sparse.copy()
    X_transformed = ohe.fit_transform(X_transformed)
    imp = SimpleImputer(copy=False)
    X_transformed = imp.fit_transform(X_transformed)
    standard_scaler = StandardScaler(with_mean=False)
    X_transformed = standard_scaler.fit_transform(X_transformed)

    # Transform the array which indicates the numerical metafeatures
    feat_type_transformed = {i: "numerical" for i in range(X_transformed.shape[1])}

    X = X_sparse
    X_transformed = X_transformed
    y = y
    mf = meta_features.metafeatures
    helpers = meta_features.helper_functions
    logger = logging.getLogger()

    # Precompute some helper functions
    helpers.set_value(
        "PCA",
        helpers["PCA"](X_transformed, y, logger),
    )
    helpers.set_value(
        "MissingValues",
        helpers["MissingValues"](X, y, logger, feat_type),
    )
    mf.set_value(
        "NumberOfMissingValues",
        mf["NumberOfMissingValues"](X, y, logger, feat_type),
    )
    helpers.set_value(
        "NumSymbols",
        helpers["NumSymbols"](X, y, logger, feat_type),
    )
    helpers.set_value(
        "ClassOccurences",
        helpers["ClassOccurences"](X, y, logger),
    )
    helpers.set_value(
        "Skewnesses",
        helpers["Skewnesses"](X_transformed, y, logger, feat_type_transformed),
    )
    helpers.set_value(
        "Kurtosisses",
        helpers["Kurtosisses"](X_transformed, y, logger, feat_type_transformed),
    )
    return X_transformed, y, feat_type_transformed


def test_missing_values(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.helper_functions["MissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert sparse.issparse(mf.value)
    assert mf.value.shape == X.shape
    assert mf.value.dtype == bool
    assert 0 == np.sum(mf.value.data)


def test_number_of_missing_values(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["NumberOfMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert 0 == mf.value


def test_percentage_missing_values(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["PercentageOfMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert 0 == mf.value


def test_number_of_Instances_with_missing_values(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["NumberOfInstancesWithMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert 0 == mf.value


def test_percentage_of_Instances_with_missing_values(sparse_data):
    X, y, categorical = sparse_data
    meta_features.metafeatures.set_value(
        "NumberOfInstancesWithMissingValues",
        meta_features.metafeatures["NumberOfInstancesWithMissingValues"](
            X, y, logging.getLogger("Meta"), categorical
        ),
    )
    mf = meta_features.metafeatures["PercentageOfInstancesWithMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert pytest.approx(0) == mf.value


def test_number_of_features_with_missing_values(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["NumberOfFeaturesWithMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert 0 == mf.value


def test_percentage_of_features_with_missing_values(sparse_data):
    X, y, categorical = sparse_data
    meta_features.metafeatures.set_value(
        "NumberOfFeaturesWithMissingValues",
        meta_features.metafeatures["NumberOfFeaturesWithMissingValues"](
            X, y, logging.getLogger("Meta"), categorical
        ),
    )
    mf = meta_features.metafeatures["PercentageOfFeaturesWithMissingValues"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert mf.value == pytest.approx(0.0)


def test_num_symbols(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.helper_functions["NumSymbols"](
        X, y, logging.getLogger("Meta"), categorical
    )

    symbol_frequency = [
        2,
        0,
        6,
        0,
        1,
        3,
        0,
        0,
        3,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        0,
        1,
        1,
        2,
        2,
    ]
    assert mf.value == symbol_frequency


def test_symbols_max(sparse_data):
    X, y, categorical = sparse_data
    # this is attribute steel
    mf = meta_features.metafeatures["SymbolsMax"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert mf.value == 6


def test_symbols_mean(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["SymbolsMean"](
        X, y, logging.getLogger("Meta"), categorical
    )
    # Empty looking spaces denote empty attributes
    symbol_frequency = [2, 6, 1, 3, 3, 1, 1, 2, 1, 1, 2, 2]
    assert pytest.approx(mf.value) == np.mean(symbol_frequency)


def test_symbols_std(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["SymbolsSTD"](
        X, y, logging.getLogger("Meta"), categorical
    )
    symbol_frequency = [2, 6, 1, 3, 3, 1, 1, 2, 1, 1, 2, 2]
    assert pytest.approx(mf.value) == np.std(symbol_frequency)


def test_symbols_sum(sparse_data):
    X, y, categorical = sparse_data
    mf = meta_features.metafeatures["SymbolsSum"](
        X, y, logging.getLogger("Meta"), categorical
    )
    assert mf.value == 25


def test_skewnesses(sparse_data_transformed):
    X_transformed, y, categorical_transformed = sparse_data_transformed
    fixture = [
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        -0.696970849903357,
        0.626346013011262,
        0.38099875966240554,
        1.4762248835141032,
        0.07687661087633788,
        0.3688979783036015,
    ]
    mf = meta_features.helper_functions["Skewnesses"](
        X_transformed, y, logging.getLogger("Meta")
    )
    np.testing.assert_allclose(mf.value, fixture)


def test_kurtosisses(sparse_data_transformed):
    fixture = [
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        -1.1005836114255763,
        -1.1786325509475744,
        -1.23879983823279,
        1.3934382644137013,
        -0.9768209837948336,
        -1.7937072296512784,
    ]
    X_transformed, y, categorical_transformed = sparse_data_transformed
    mf = meta_features.helper_functions["Kurtosisses"](
        X_transformed, y, logging.getLogger("Meta")
    )
    np.testing.assert_allclose(mf.value, fixture)


def test_pca_95percent(sparse_data_transformed):
    X_transformed, y, categorical_transformed = sparse_data_transformed
    mf = meta_features.metafeatures["PCAFractionOfComponentsFor95PercentVariance"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(0.7741935483870968) == mf.value


def test_pca_kurtosis_first_pc(sparse_data_transformed):
    X_transformed, y, categorical_transformed = sparse_data_transformed
    mf = meta_features.metafeatures["PCAKurtosisFirstPC"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(-0.15444516166802469) == mf.value


def test_pca_skewness_first_pc(sparse_data_transformed):
    X_transformed, y, categorical_transformed = sparse_data_transformed
    mf = meta_features.metafeatures["PCASkewnessFirstPC"](
        X_transformed, y, logging.getLogger("Meta")
    )
    assert pytest.approx(0.026514792083623905) == mf.value


def test_calculate_all_metafeatures(sparse_data):
    X, y, feat_type = sparse_data
    mf = meta_features.calculate_all_metafeatures(
        X, y, feat_type, "2", logger=logging.getLogger("Meta")
    )
    assert 52 == len(mf.metafeature_values)



from typing import Any, Dict, List, Union

import warnings
from itertools import chain

import numpy as np
import pandas as pd
import sklearn.datasets
from scipy.sparse import csr_matrix, spmatrix

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    CLASSIFICATION_TASKS,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
    REGRESSION_TASKS,
)
from autosklearn.util.data import (
    default_dataset_compression_arg,
    reduce_dataset_size_if_too_large,
    reduce_precision,
    reduction_mapping,
    subsample,
    supported_precision_reductions,
    validate_dataset_compression_arg,
)

import pytest

parametrize = pytest.mark.parametrize


@parametrize("arg", [default_dataset_compression_arg])
def test_validate_dataset_compression_arg_returns_original_dict(arg: Dict[str, Any]):
    """
    Parameters
    ----------
    arg: Dict[str, any]

    Expects
    -------
    * Should validate without error
    * Should return a dict identical to the default
    """
    validate_arg = validate_dataset_compression_arg(arg, memory_limit=10)
    assert validate_arg == arg


@parametrize("memory_allocation", [10, 50, 0.1, 0.5])
def test_validate_dataset_compression_arg_returns_with_memory_allocation(
    memory_allocation: Union[float, int],
):
    """
    Parameters
    ----------
    memory_allocation: Union[float, int]
        A valid memory_allocation

    Expects
    -------
    * Should validate without error
    * Should not modify the memory_allocation arg
    * Should populate the default methods
    """
    arg = {"memory_allocation": memory_allocation}
    validate_arg = validate_dataset_compression_arg(arg, memory_limit=100)

    expected_methods = default_dataset_compression_arg["methods"]

    assert validate_arg["memory_allocation"] == memory_allocation
    assert validate_arg["methods"] == expected_methods


@parametrize(
    "methods",
    [
        ["precision"],
        ["precision", "subsample"],
        ["precision", "precision", "subsample"],
    ],
)
def test_validate_dataset_compression_arg_returns_with_same_methods(
    methods: List[str],
):
    """
    Parameters
    ----------
    methods: List[str]
        A valid list of methods

    Expects
    -------
    * Should validate without error
    * Should not modify the methods arg
    * Should populate memory_allocation with the default
    """
    arg = {"methods": methods}
    validate_arg = validate_dataset_compression_arg(arg, memory_limit=10)

    expected_memory_allocation = default_dataset_compression_arg["memory_allocation"]

    assert validate_arg["memory_allocation"] == expected_memory_allocation
    assert validate_arg["methods"] == methods


@parametrize("bad_arg", [[1337], "hello"])
def test_validate_dataset_compression_arg_raises_error_with_bad_arg(bad_arg: Any):
    """
    Parameters
    ----------
    bad_arg: Any
        An arg which is not a Mapping

    Expects
    -------
    * Should raise a ValueError
    """
    with pytest.raises(ValueError, match=r"Unknown type"):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize("key", ["hello", "hi", "hey"])
def test_validate_dataset_compression_arg_raises_error_with_bad_key(key: str):
    """
    Parameters
    ----------
    key: str
        A key which should not be in the parameters

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {**default_dataset_compression_arg, key: 1337}
    with pytest.raises(ValueError, match=r"Unknown key"):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize("memory_allocation", ["hello", {}, [1, 2, 3]])
def test_validate_dataset_compression_arg_raises_error_with_bad_memory_allocation_type(
    memory_allocation: Any,
):
    """
    Parameters
    ----------
    memory_allocation: Any
        A bad type for memory_allocation

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {"memory_allocation": memory_allocation}
    with pytest.raises(
        ValueError, match=r"key 'memory_allocation' must be an `int` or `float`"
    ):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize("memory_allocation", [-0.5, 0.0, 1.0, 1.5])
def test_validate_dataset_compression_arg_raises_error_with_bad_memory_allocation_float(
    memory_allocation: float,
):
    """
    Parameters
    ----------
    memory_allocation: Any
        A bad float value for memory_allocation

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {"memory_allocation": memory_allocation}

    with pytest.raises(
        ValueError, match=r"key 'memory_allocation' if float must be in \(0, 1\)"
    ):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize(
    "memory_allocation, memory_limit", [(0, 10), (10, 10), (-20, 10), (20, 10)]
)
def test_validate_dataset_compression_arg_raises_error_with_bad_memory_allocation_int(
    memory_allocation: int, memory_limit: int
):
    """
    Parameters
    ----------
    memory_allocation: int
        A bad float int for memory_allocation

    memory_limit: int
        The memory limit

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {"memory_allocation": memory_allocation}
    with pytest.raises(
        ValueError, match=r"key 'memory_allocation' if int must be in \(0,"
    ):
        validate_dataset_compression_arg(bad_arg, memory_limit=memory_limit)


@parametrize("methods", [10, {"hello", "world"}, []])
def test_validate_dataset_compression_arg_raises_error_with_bad_methods_type(
    methods: Any,
):
    """
    Parameters
    ----------
    methods: int
        A bad type for key methods

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {"methods": methods}
    with pytest.raises(ValueError, match=r"key 'methods' must be a non-empty list"):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize(
    "methods",
    [
        ["bad", "worse"],
        ["precision", "kind_of_bad"],
        ["still_bad", "precision", "subsample"],
    ],
)
def test_validate_dataset_compression_arg_raises_error_with_bad_methods_entries(
    methods: Any,
):
    """
    Parameters
    ----------
    methods: int
        A bad type for key methods

    Expects
    -------
    * Should raise a ValueError
    """
    bad_arg = {"methods": methods}
    with pytest.raises(ValueError, match=r"key 'methods' can only contain"):
        validate_dataset_compression_arg(bad_arg, memory_limit=10)


@parametrize(
    "y",
    [
        np.asarray(9999 * [0] + 1 * [1]),
        np.asarray(4999 * [1] + 4999 * [2] + 1 * [3] + 1 * [4]),
        np.asarray(
            4999 * [[0, 1, 1]] + 4999 * [[1, 1, 0]] + 1 * [[1, 0, 1]] + 1 * [[0, 0, 0]]
        ),
    ],
)
@parametrize("random_state", list(range(5)))
def test_subsample_classification_unique_labels_stay_in_training_set(y, random_state):
    n_samples = len(y)
    X = np.random.random(size=(n_samples, 3))
    sample_size = 100

    values, counts = np.unique(y, axis=0, return_counts=True)
    unique_labels = [value for value, count in zip(values, counts) if count == 1]
    assert len(unique_labels), "Ensure we have unique labels in the test"

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X_sampled, y_sampled = subsample(
            X,
            y,
            random_state=random_state,
            sample_size=sample_size,
            is_classification=True,
        )

    assert X_sampled.dtype == X.dtype and y_sampled.dtype == y.dtype
    assert len(y_sampled) == sample_size
    assert all(
        label in y_sampled for label in unique_labels
    ), f"sampled unique = {np.unique(y_sampled)}, original unique = {unique_labels}"


@parametrize("X", [np.asarray([[1, 1, 1]] * 30)])
@parametrize("x_type", [list, np.ndarray, csr_matrix, pd.DataFrame])
@parametrize(
    "y, task",
    [
        (np.asarray([0] * 15 + [1] * 15), BINARY_CLASSIFICATION),
        (np.asarray([0] * 10 + [1] * 10 + [2] * 10), MULTICLASS_CLASSIFICATION),
        (np.asarray([[1, 0, 1]] * 30), MULTILABEL_CLASSIFICATION),
        (np.asarray([1.0] * 30), REGRESSION),
        (np.asarray([[1.0, 1.0, 1.0]] * 30), MULTIOUTPUT_REGRESSION),
    ],
)
@parametrize("y_type", [list, np.ndarray, pd.DataFrame, pd.Series])
@parametrize("random_state", [0])
@parametrize("sample_size", [0.25, 0.5, 5, 10])
def test_subsample_validity(X, x_type, y, y_type, random_state, sample_size, task):
    """Asserts the validity of the function with all valid types

    We want to make sure that `subsample` works correctly with all the types listed
    as x_type and y_type.

    We also want to make sure it works with all kinds of target types.

    The output should maintain the types, and subsample the correct amount.
    """
    assert len(X) == len(y)  # Make sure our test data is correct

    if y_type == pd.Series and task in [
        MULTILABEL_CLASSIFICATION,
        MULTIOUTPUT_REGRESSION,
    ]:
        # We can't have a pd.Series with multiple values as it's 1 dimensional
        pytest.skip("Can't have pd.Series as y when task is n-dimensional")

    # Convert our data to its given x_type or y_type
    def convert(arr, objtype):
        if objtype == np.ndarray:
            return arr
        elif objtype == list:
            return arr.tolist()
        else:
            return objtype(arr)

    X = convert(X, x_type)
    y = convert(y, y_type)

    # Subsample the data, ignoring any warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X_sampled, y_sampled = subsample(
            X,
            y,
            random_state=random_state,
            sample_size=sample_size,
            is_classification=task in CLASSIFICATION_TASKS,
        )

    # Function to get the type of an obj
    def dtype(obj):
        if isinstance(obj, List):
            if isinstance(obj[0], List):
                return type(obj[0][0])
            else:
                return type(obj[0])

        elif isinstance(obj, pd.DataFrame):
            return obj.dtypes

        else:
            return obj.dtype

    # Check that the types of X remain the same after subsampling
    if isinstance(X, pd.DataFrame):
        # Dataframe can have multiple types, one per column
        assert list(dtype(X_sampled)) == list(dtype(X))
    else:
        assert dtype(X_sampled) == dtype(X)

    # Check that the types of y remain the same after subsampling
    if isinstance(y, pd.DataFrame):
        assert list(dtype(y_sampled)) == list(dtype(y))
    else:
        assert dtype(y_sampled) == dtype(y)

    # Function to get the size of an object
    def size(obj):
        if isinstance(obj, spmatrix):  # spmatrix doesn't support __len__
            return obj.shape[0] if obj.shape[0] > 1 else obj.shape[1]
        else:
            return len(obj)

    # check the right amount of samples were taken
    if sample_size < 1:
        assert size(X_sampled) == int(sample_size * size(X))
    else:
        assert size(X_sampled) == sample_size


@parametrize("X", [np.asarray([[0, 0, 1]] * 10)])
@parametrize(
    "dtype", supported_precision_reductions + [np.dtype("float32"), np.dtype("float64")]
)
@parametrize("x_type", [np.ndarray, csr_matrix])
def test_reduce_precision_correctly_reduces_precision(X, dtype, x_type):
    X = X.astype(dtype)
    if x_type == csr_matrix:
        X = x_type(X)

    X_reduced, precision = reduce_precision(X)

    # Check the reduced precision is correctly returned
    assert X_reduced.dtype == precision

    # Check that it was reduce to the correct precision
    expected: Dict[type, type] = {
        np.float32: np.float32,
        np.float64: np.float32,
        np.dtype("float32"): np.float32,
        np.dtype("float64"): np.float32,
    }
    if hasattr(np, "float96"):
        expected[np.float96] = np.float64

    if hasattr(np, "float128"):
        expected[np.float128] = np.float64

    assert precision == expected[dtype]

    # Check that X's shape was not modified in any way
    assert X.shape == X_reduced.shape

    # Check that the return type is the one as we passed in
    assert type(X) == type(X_reduced)


@parametrize("X", [np.asarray([0, 0, 1]) * 10])
@parametrize("dtype", [np.int32, np.int64, np.complex128])
def test_reduce_precision_with_unsupported_dtypes(X, dtype):
    X = X.astype(dtype)
    with pytest.raises(ValueError) as err:
        reduce_precision(X)

    expected = (
        f"X.dtype = {X.dtype} not equal to any supported "
        f"{supported_precision_reductions}"
    )

    assert err.value.args[0] == expected


@parametrize(
    "X",
    [
        np.ones(
            (100000, 10), dtype=np.float64
        )  # Make it big for reductions to take place
    ],
)
@parametrize("x_type", [csr_matrix, np.ndarray])
@parametrize("dtype", supported_precision_reductions)
@parametrize(
    "y, is_classification",
    [
        (np.ones((100000,)), True),
        (np.ones((100000,)), False),
    ],
)
@parametrize("memory_allocation", [0.1, 1 / 5.2, 1 / 8, 1])
@parametrize("operations", [["precision"], ["subsample"], ["precision", "subsample"]])
def test_reduce_dataset_reduces_size_and_precision(
    X, x_type, dtype, y, is_classification, memory_allocation, operations
):
    assert len(X) == len(y)
    X = X.astype(dtype)
    if x_type == csr_matrix:
        X = x_type(X)

    random_state = 0
    memory_limit = 2  # Force reductions

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        X_out, y_out = reduce_dataset_size_if_too_large(
            X=X,
            y=y,
            random_state=random_state,
            memory_limit=memory_limit,
            operations=operations,
            memory_allocation=memory_allocation,
            is_classification=is_classification,
        )

    def bytes(arr):
        return arr.nbytes if isinstance(arr, np.ndarray) else arr.data.nbytes

    # If we expect some precision reduction unless at float32 already
    if "precision" in operations and dtype != np.float32:
        expected = reduction_mapping[X.dtype]
        assert X_out.dtype == expected
        assert bytes(X_out) < bytes(X)

    # If we expect some subsampling
    if "subsample" in operations:
        assert X_out.shape[0] < X.shape[0]
        assert y_out.shape[0] < y.shape[0]
        assert bytes(X_out) < bytes(X)


def test_reduce_dataset_invalid_dtype_for_precision_reduction():
    dtype = int
    X = np.asarray([1, 2, 3], dtype=dtype)

    with pytest.raises(ValueError) as err:
        reduce_dataset_size_if_too_large(
            X=X,
            y=X,
            operations=["precision"],
            memory_limit=1,
            memory_allocation=0.1,
            is_classification=False,
        )

    expected_err = f"Unsupported type `{X.dtype}` for precision reduction"
    assert err.value.args[0] == expected_err


def test_reduce_dataset_invalid_operations():
    invalid_op = "invalid"

    X = np.asarray([1, 2, 3], dtype=float)
    with pytest.raises(ValueError) as err:
        reduce_dataset_size_if_too_large(
            X=X,
            y=X,
            operations=[invalid_op],
            memory_limit=1,
            memory_allocation=0.1,
            is_classification=False,
        )

    expected_err = f"Unknown operation `{invalid_op}`"
    assert err.value.args[0] == expected_err


@parametrize("memory_allocation", [-0.5, 0.0, 1.0, 1.5])
def test_reduce_dataset_invalid_memory_allocation_float(memory_allocation: float):
    """
    Parameters
    ----------
    memory_allocation: float
        An invalid memory allocation as a float

    Expects
    -------
    * Should raise a ValueError
    """
    with pytest.raises(
        ValueError, match=r"memory_allocation if float must be in \(0, 1\)"
    ):
        reduce_dataset_size_if_too_large(
            X=np.empty(1),
            y=np.empty(1),
            memory_limit=100,
            is_classification=True,
            memory_allocation=memory_allocation,
        )


@parametrize("memory_allocation", [-1, 0, 100, 101])
def test_reduce_dataset_invalid_memory_allocation_int(memory_allocation: int):
    """
    Parameters
    ----------
    memory_allocation: float
        An invalid memory allocation as a int

    Expects
    -------
    * Should raise a ValueError
    """
    with pytest.raises(
        ValueError, match=r"memory_allocation if int must be in \(0, memory_limit"
    ):
        reduce_dataset_size_if_too_large(
            X=np.empty(1),
            y=np.empty(1),
            is_classification=True,
            memory_limit=100,
            memory_allocation=memory_allocation,
        )


@parametrize("memory_allocation", ["100", {"a": 1}, [100]])
def test_reduce_dataset_invalid_memory_allocation_type(memory_allocation: Any):
    """
    Parameters
    ----------
    memory_allocation: Any
        An invalid memory allocation type

    Expects
    -------
    * Should raise a ValueError
    """
    with pytest.raises(ValueError, match=r"Unknown type for `memory_allocation`"):
        reduce_dataset_size_if_too_large(
            X=np.empty(1),
            y=np.empty(1),
            memory_limit=100,
            is_classification=True,
            memory_allocation=memory_allocation,
        )


@pytest.mark.parametrize(
    "memory_limit,precision,task",
    [
        (memory_limit, precision, task)
        for task in chain(CLASSIFICATION_TASKS, REGRESSION_TASKS)
        for precision in (float, np.float32, np.float64, np.float128)
        for memory_limit in (1, 100)
    ],
)
def test_reduce_dataset_subsampling_explicit_values(memory_limit, precision, task):
    random_state = 0
    fixture = {
        BINARY_CLASSIFICATION: {
            1: {float: 2621, np.float32: 2621, np.float64: 2621, np.float128: 1310},
            100: {
                float: 12000,
                np.float32: 12000,
                np.float64: 12000,
                np.float128: 12000,
            },
        },
        MULTICLASS_CLASSIFICATION: {
            1: {float: 409, np.float32: 409, np.float64: 409, np.float128: 204},
            100: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
        },
        MULTILABEL_CLASSIFICATION: {
            1: {float: 409, np.float32: 409, np.float64: 409, np.float128: 204},
            100: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
        },
        REGRESSION: {
            1: {float: 1310, np.float32: 1310, np.float64: 1310, np.float128: 655},
            100: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
        },
        MULTIOUTPUT_REGRESSION: {
            1: {float: 1310, np.float32: 1310, np.float64: 1310, np.float128: 655},
            100: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
        },
    }

    # Create the task and data
    if task == BINARY_CLASSIFICATION:
        X, y = sklearn.datasets.make_hastie_10_2()
    elif task == MULTICLASS_CLASSIFICATION:
        X, y = sklearn.datasets.load_digits(return_X_y=True)
    elif task == MULTILABEL_CLASSIFICATION:
        X, y_ = sklearn.datasets.load_digits(return_X_y=True)
        y = np.zeros((X.shape[0], 10))
        for i, j in enumerate(y_):
            y[i, j] = 1
    elif task == REGRESSION:
        X, y = sklearn.datasets.make_friedman1(n_samples=5000, n_features=20)
    elif task == MULTIOUTPUT_REGRESSION:
        X, y = sklearn.datasets.make_friedman1(n_samples=5000, n_features=20)
        y = np.vstack((y, y)).transpose()
    else:
        raise ValueError(task)

    # Validate the test data and make sure X and y have the same number of rows
    assert X.shape[0] == y.shape[0]

    # Convert X to the dtype we are testing
    X = X.astype(precision)

    # Preform the subsampling through `reduce_dataset_size_if_too_large`
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X_new, y_new = reduce_dataset_size_if_too_large(
            X=X,
            y=y,
            random_state=random_state,
            memory_limit=memory_limit,
            is_classification=task in CLASSIFICATION_TASKS,
            operations=["precision", "subsample"],
            memory_allocation=0.1,
        )

    # Assert the new number of samples
    assert X_new.shape[0] == fixture[task][memory_limit][precision]



import multiprocessing
import sys


def preload_modules(context: multiprocessing.context.BaseContext) -> None:
    all_loaded_modules = sys.modules.keys()
    preload = [
        loaded_module
        for loaded_module in all_loaded_modules
        if loaded_module.split(".")[0]
        in (
            "smac",
            "autosklearn",
            "numpy",
            "scipy",
            "pandas",
            "pynisher",
            "sklearn",
            "ConfigSpace",
        )
        and "logging" not in loaded_module
    ]
    context.set_forkserver_preload(preload)



import numpy as np

from autosklearn.pipeline.components.feature_preprocessing.select_percentile_regression import (  # noqa: E501
    SelectPercentileRegression,
)
from autosklearn.pipeline.util import _test_preprocessing, get_dataset

import unittest


class SelectPercentileRegressionTest(unittest.TestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(
            dataset="boston",
            Preprocessor=SelectPercentileRegression,
        )
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], int(original.shape[1] / 2))
        self.assertFalse((transformation == 0).all())

    def test_preprocessing_dtype(self):
        # Dense
        # np.float32
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        self.assertEqual(X_train.dtype, np.float32)

        configuration_space = (
            SelectPercentileRegression.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileRegression(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float32)

        # np.float64
        X_train, Y_train, X_test, Y_test = get_dataset("iris")
        X_train = X_train.astype(np.float64)
        configuration_space = (
            SelectPercentileRegression.get_hyperparameter_search_space()
        )
        default = configuration_space.get_default_configuration()
        preprocessor = SelectPercentileRegression(
            random_state=1, **{hp_name: default[hp_name] for hp_name in default}
        )
        preprocessor.fit(X_train, Y_train)
        Xt = preprocessor.transform(X_train)
        self.assertEqual(Xt.dtype, np.float64)



import sklearn.metrics
from sklearn.linear_model import RidgeClassifier

from autosklearn.pipeline.components.feature_preprocessing.liblinear_svc_preprocessor import (  # noqa: E501
    LibLinear_Preprocessor,
)
from autosklearn.pipeline.util import (
    PreprocessingTestCase,
    _test_preprocessing,
    get_dataset,
)

from test.test_pipeline.ignored_warnings import (
    feature_preprocessing_warnings,
    ignore_warnings,
)


class LiblinearComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        with ignore_warnings(feature_preprocessing_warnings):
            transformation, original = _test_preprocessing(LibLinear_Preprocessor)

        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertFalse((transformation == 0).all())

    def test_default_configuration_classify(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(
                dataset="digits", make_sparse=False
            )
            configuration_space = (
                LibLinear_Preprocessor.get_hyperparameter_search_space()
            )
            default = configuration_space.get_default_configuration()
            preprocessor = LibLinear_Preprocessor(
                random_state=1,
                **{
                    hp_name: default[hp_name]
                    for hp_name in default
                    if default[hp_name] is not None
                },
            )

            with ignore_warnings(feature_preprocessing_warnings):
                preprocessor.fit(X_train, Y_train)

            X_train_trans = preprocessor.transform(X_train)
            X_test_trans = preprocessor.transform(X_test)

            # fit a classifier on top
            classifier = RidgeClassifier()
            predictor = classifier.fit(X_train_trans, Y_train)
            predictions = predictor.predict(X_test_trans)
            accuracy = sklearn.metrics.accuracy_score(predictions, Y_test)
            self.assertAlmostEqual(accuracy, 0.8548876745598057, places=2)

    def test_preprocessing_dtype(self):

        with ignore_warnings(feature_preprocessing_warnings):
            super()._test_preprocessing_dtype(LibLinear_Preprocessor, test_sparse=False)



"""
Here we define the several different setups and cache them to allow for easier unit
testing. The caching mechanism is only per session and does not persist over sessions.
There's only really a point for caching fitted models that will be tested multiple times
for different properties.

Anything using a cached model must not destroy any backend resources, although it can
mock if required.

Tags:
    {classifier, regressor} - The type of AutoML object
        classifier - will be fit on "iris"
        regressor - will be fit on "boston"
    {fitted} - If the automl case has been fitted
    {cv, holdout} - Whether explicitly cv or holdout was used
    {no_ensemble} - Fit with no ensemble size
    {multiobjective} - If the automl instance is multiobjective
"""
from __future__ import annotations

from typing import Callable, Tuple

from pathlib import Path

import numpy as np
import sklearn.model_selection

import autosklearn.metrics
from autosklearn.automl import AutoMLClassifier, AutoMLRegressor
from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.evaluation.abstract_evaluator import (
    MyDummyClassifier,
    MyDummyRegressor,
)

from pytest_cases import case, parametrize

from test.fixtures.backend import copy_backend
from test.fixtures.caching import Cache


def stop_at_first(smbo, run_info, result, time_left) -> bool:
    """Used in some cases to enforce the only valid model is the dummy model"""
    return False


@case(tags=["classifier"])
def case_classifier(
    tmp_dir: str,
    make_automl_classifier: Callable[..., AutoMLClassifier],
) -> AutoMLClassifier:
    """Case basic unfitted AutoMLClassifier"""
    dir = Path(tmp_dir) / "backend"
    model = make_automl_classifier(temporary_directory=str(dir))
    return model


@case(tags=["classifier"])
def case_regressor(
    tmp_dir: str,
    make_automl_regressor: Callable[..., AutoMLRegressor],
) -> AutoMLRegressor:
    """Case basic unfitted AutoMLClassifier"""
    dir = Path(tmp_dir) / "backend"
    model = make_automl_regressor(temporary_directory=str(dir))
    return model


# ###################################
# The following are fitted and cached
# ###################################
@case(tags=["classifier", "fitted", "holdout"])
@parametrize("dataset", ["iris"])
def case_classifier_fitted_holdout_iterative(
    dataset: str,
    make_cache: Callable[[str], Cache],
    make_backend: Callable[..., Backend],
    make_automl_classifier: Callable[..., AutoMLClassifier],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLClassifier:
    """Case of a holdout fitted classifier"""
    resampling_strategy = "holdout-iterative-fit"

    key = f"case_classifier_{resampling_strategy}_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:
        if "model" not in cache:
            # Make the model in the cache
            model = make_automl_classifier(
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                resampling_strategy=resampling_strategy,
            )

            X, y, Xt, yt = make_sklearn_dataset(name=dataset)
            model.fit(X, y, dataset_name=dataset)

            # Save the model
            cache.save(model, "model")

    # Try the model from the cache
    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["classifier", "fitted", "cv"])
@parametrize("dataset", ["iris"])
def case_classifier_fitted_cv(
    make_cache: Callable[[str], Cache],
    dataset: str,
    make_backend: Callable[..., Backend],
    make_automl_classifier: Callable[..., AutoMLClassifier],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLClassifier:
    """Case of a fitted cv AutoMLClassifier"""
    resampling_strategy = "cv"

    key = f"case_classifier_{resampling_strategy}_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:
        if "model" not in cache:
            model = make_automl_classifier(
                resampling_strategy=resampling_strategy,
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                time_left_for_this_task=60,  # Give some more for CV
                per_run_time_limit=10,
            )

            X, y, Xt, yt = make_sklearn_dataset(name=dataset)
            model.fit(X, y, dataset_name=dataset)

            cache.save(model, "model")

    # Try the model from the cache
    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["classifier", "fitted", "holdout", "multiobjective"])
@parametrize("dataset", ["iris"])
def case_classifier_fitted_holdout_multiobjective(
    dataset: str,
    make_cache: Callable[[str], Cache],
    make_backend: Callable[..., Backend],
    make_automl_classifier: Callable[..., AutoMLClassifier],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLClassifier:
    """Case of a holdout fitted classifier"""
    resampling_strategy = "holdout"

    key = f"case_classifier_{resampling_strategy}_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:
        if "model" not in cache:
            # Make the model in the cache
            model = make_automl_classifier(
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                resampling_strategy=resampling_strategy,
                metrics=[
                    autosklearn.metrics.balanced_accuracy,
                    autosklearn.metrics.log_loss,
                ],
            )

            X, y, Xt, yt = make_sklearn_dataset(
                name=dataset, return_target_as_string=True
            )
            model.fit(X, y, dataset_name=dataset)

            # Save the model
            cache.save(model, "model")

    # Try the model from the cache
    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["regressor", "fitted", "holdout"])
@parametrize("dataset", ["boston"])
def case_regressor_fitted_holdout(
    make_cache: Callable[[str], Cache],
    dataset: str,
    make_backend: Callable[..., Backend],
    make_automl_regressor: Callable[..., AutoMLRegressor],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLRegressor:
    """Case of fitted regressor with holdout"""
    resampling_strategy = "holdout"

    key = f"case_regressor_{resampling_strategy}_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:
        if "model" not in cache:
            model = make_automl_regressor(
                temporary_directory=cache.path("backend"),
                resampling_strategy=resampling_strategy,
                delete_tmp_folder_after_terminate=False,
            )

            X, y, Xt, yt = make_sklearn_dataset(name=dataset)
            model.fit(X, y, dataset_name=dataset)

            cache.save(model, "model")

    # Try the model from the cache
    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["regressor", "fitted", "cv"])
@parametrize("dataset", ["boston"])
def case_regressor_fitted_cv(
    make_cache: Callable[[str], Cache],
    dataset: str,
    make_backend: Callable[..., Backend],
    make_automl_regressor: Callable[..., AutoMLRegressor],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLRegressor:
    """Case of fitted regressor with cv resampling"""
    resampling_strategy = "cv"
    key = f"case_regressor_{resampling_strategy}_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:

        if "model" not in cache:
            model = make_automl_regressor(
                temporary_directory=cache.path("backend"),
                resampling_strategy=resampling_strategy,
                delete_tmp_folder_after_terminate=False,
                time_left_for_this_task=60,  # Some extra time for CV
                per_run_time_limit=10,
            )

            X, y, Xt, yt = make_sklearn_dataset(name=dataset)
            model.fit(X, y, dataset_name=dataset)

            cache.save(model, "model")

    # Try the model from the cache
    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["classifier", "fitted", "no_ensemble"])
@parametrize("dataset", ["iris"])
def case_classifier_fitted_no_ensemble(
    make_cache: Callable[[str], Cache],
    dataset: str,
    make_backend: Callable[..., Backend],
    make_automl_classifier: Callable[..., AutoMLClassifier],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> AutoMLClassifier:
    """Case of a fitted classifier but ensemble was disabled"""
    key = f"case_classifier_fitted_no_ensemble_{dataset}"

    # This locks the cache for this item while we check, required for pytest-xdist
    with make_cache(key) as cache:

        if "model" not in cache:
            model = make_automl_classifier(
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                ensemble_class=None,
            )

            X, y, Xt, yt = make_sklearn_dataset(name=dataset)
            model.fit(X, y, dataset_name=dataset)

            cache.save(model, "model")

    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["classifier", "fitted"])
def case_classifier_fitted_only_dummy(
    make_cache: Callable[[str], Cache],
    make_backend: Callable[..., Backend],
    make_automl_classifier: Callable[..., AutoMLClassifier],
) -> AutoMLClassifier:
    """Case of a fitted classifier but only dummy was found"""
    key = "case_classifier_fitted_only_dummy"

    # This locks the cache for this item while we check, required for pytest-xdist

    with make_cache(key) as cache:
        if "model" not in cache:
            model = make_automl_classifier(
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                include={"classifier": ["bernoulli_nb"]},  # Just a meh model
                get_trials_callback=stop_at_first,
            )
            rand = np.random.RandomState(2)
            _X = rand.random((100, 50))
            _y = rand.randint(0, 2, (100,))
            X, Xt, y, yt = sklearn.model_selection.train_test_split(
                _X, _y, random_state=1  # Required to ensure dummy is best
            )
            model.fit(X, y, dataset_name="random")

            # We now validate that indeed, the only model is the Dummy
            members = list(model.models_.values())
            if len(members) != 1 and not isinstance(members[0], MyDummyClassifier):
                raise ValueError("Should only have one model, dummy\n", members)

            cache.save(model, "model")

    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model


@case(tags=["regressor", "fitted"])
def case_regressor_fitted_only_dummy(
    make_cache: Callable[[str], Cache],
    make_backend: Callable[..., Backend],
    make_automl_regressor: Callable[..., AutoMLRegressor],
) -> AutoMLRegressor:
    """Case of a fitted classifier but only dummy was found"""
    key = "case_regressor_fitted_only_dummy"

    # This locks the cache for this item while we check, required for pytest-xdist

    with make_cache(key) as cache:
        if "model" not in cache:
            model = make_automl_regressor(
                temporary_directory=cache.path("backend"),
                delete_tmp_folder_after_terminate=False,
                include={"regressor": ["k_nearest_neighbors"]},  # Just a meh model
                get_trials_callback=stop_at_first,
            )

            rand = np.random.RandomState(2)
            _X = rand.random((100, 50))
            _y = rand.random((100,))

            X, Xt, y, yt = sklearn.model_selection.train_test_split(
                _X, _y, random_state=1  # Required to ensure dummy is best
            )
            model.fit(X, y, dataset_name="random")

            # We now validate that indeed, the only model is the Dummy
            members = list(model.models_.values())
            if len(members) != 1 and not isinstance(members[0], MyDummyRegressor):
                raise ValueError("Should only have one model, dummy\n", members)

            cache.save(model, "model")

    model = cache.load("model")
    model._backend = copy_backend(old=model._backend, new=make_backend())

    return model



import logging
import os

import pandas as pd

import autosklearn.pipeline.classification
from autosklearn.metalearning.metalearning.meta_base import MetaBase

import unittest


class MetaBaseTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.cwd = os.getcwd()
        data_dir = os.path.dirname(__file__)
        data_dir = os.path.join(data_dir, "test_meta_base_data")
        os.chdir(data_dir)

        pipeline = autosklearn.pipeline.classification.SimpleClassificationPipeline()
        cs = pipeline.get_hyperparameter_search_space()

        self.logger = logging.getLogger()
        self.base = MetaBase(cs, data_dir, logger=self.logger)

    def tearDown(self):
        os.chdir(self.cwd)

    def test_get_all_runs(self):
        runs = self.base.get_all_runs()
        self.assertIsInstance(runs, pd.DataFrame)
        # TODO update this ASAP
        self.assertEqual((125, 125), runs.shape)

    def test_get_runs(self):
        runs = self.base.get_runs("233")
        # TODO update this ASAP
        self.assertEqual(125, len(runs))
        self.assertIsInstance(runs, pd.Series)

    def test_get_metafeatures_single_dataset(self):
        mf = self.base.get_metafeatures("233")
        self.assertIsInstance(mf, pd.Series)
        self.assertEqual(mf.name, "233")
        self.assertEqual(mf.loc["NumberOfInstances"], 2142.0)

    def test_get_metafeatures_single_feature(self):
        mf = self.base.get_metafeatures(features="NumberOfInstances")
        self.assertIsInstance(mf, pd.Series)
        self.assertEqual(mf.shape, (132,))

    def test_get_metafeatures_single_dataset_and_single_feature(self):
        mf = self.base.get_metafeatures("233", features="NumberOfInstances")
        self.assertEqual(mf.shape, ())

    def test_get_metafeatures_multiple_datasets(self):
        mf = self.base.get_metafeatures(["233", "236"])
        self.assertIsInstance(mf, pd.DataFrame)
        self.assertEqual(mf.shape, (2, 46))

    def test_get_metafeatures_multiple_features(self):
        mf = self.base.get_metafeatures(
            features=["NumberOfInstances", "NumberOfClasses"]
        )
        self.assertIsInstance(mf, pd.DataFrame)
        self.assertEqual(mf.shape, (132, 2))



from collections import OrderedDict

from autosklearn.metalearning.optimizers import optimizer_base

import unittest


class OptimizerBaseTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.hyperparameters = OrderedDict()
        self.hyperparameters["x"] = [-5, 0, 5, 10]
        self.hyperparameters["y"] = [0, 5, 10, 15]

    def test_parse_hyperopt_string(self):
        hyperparameter_string = "x {-5, 0, 5, 10}\ny {0, 5, 10, 15}"
        expected = OrderedDict(
            [["x", ["-5", "0", "5", "10"]], ["y", ["0", "5", "10", "15"]]]
        )
        ret = optimizer_base.parse_hyperparameter_string(hyperparameter_string)
        self.assertEqual(ret, expected)

        hyperparameter_string = "x {-5, 0, 5, 10} [5]\ny {0, 5, 10, 15}"
        ret = optimizer_base.parse_hyperparameter_string(hyperparameter_string)
        self.assertEqual(ret, expected)

        hyperparameter_string = "x {-5, 0, 5, 10}\ny {0, 5, 10, 15} [5]"
        ret = optimizer_base.parse_hyperparameter_string(hyperparameter_string)
        self.assertEqual(ret, expected)

        hyperparameter_string = "x {-5, 0, 5, 10}\ny 0, 5, 10, 15} [5]"
        self.assertRaises(
            ValueError,
            optimizer_base.parse_hyperparameter_string,
            hyperparameter_string,
        )

    def test_construct_cli_call(self):
        cli_call = optimizer_base.construct_cli_call("cv.py", {"x": -5, "y": 0})
        self.assertEqual(cli_call, "cv.py -x \"'-5'\" -y \"'0'\"")



# -*- encoding: utf-8 -*-
from sklearn.datasets import load_breast_cancer

from autosklearn.classification import AutoSklearnClassifier
from autosklearn.constants import MULTICLASS_CLASSIFICATION, REGRESSION
from autosklearn.metalearning.mismbo import suggest_via_metalearning
from autosklearn.pipeline.util import get_dataset
from autosklearn.smbo import _calculate_metafeatures, _calculate_metafeatures_encoded
from autosklearn.util.pipeline import get_configuration_space

import unittest


class MetafeatureValueDummy(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value


class Test(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.X_train, self.Y_train, self.X_test, self.Y_test = get_dataset("iris")

        eliminate_class_two = self.Y_train != 2
        self.X_train = self.X_train[eliminate_class_two]
        self.Y_train = self.Y_train[eliminate_class_two]

    @unittest.skip("TODO refactor!")
    def test_metalearning(self):
        dataset_name_classification = "digits"
        initial_challengers_classification = {
            "ACC_METRIC": '--initial-challengers " '
            "-balancing:strategy 'weighting' "
            "-classifier:__choice__ 'proj_logit'",
            "AUC_METRIC": '--initial-challengers " '
            "-balancing:strategy 'weighting' "
            "-classifier:__choice__ 'liblinear_svc'",
            "BAC_METRIC": '--initial-challengers " '
            "-balancing:strategy 'weighting' "
            "-classifier:__choice__ 'proj_logit'",
            "F1_METRIC": '--initial-challengers " '
            "-balancing:strategy 'weighting' "
            "-classifier:__choice__ 'proj_logit'",
            "PAC_METRIC": '--initial-challengers " '
            "-balancing:strategy 'none' "
            "-classifier:__choice__ 'random_forest'",
        }

        dataset_name_regression = "diabetes"
        initial_challengers_regression = {
            "A_METRIC": '--initial-challengers " '
            "-imputation:strategy 'mean' "
            "-one_hot_encoding:minimum_fraction '0.01' "
            "-one_hot_encoding:use_minimum_fraction 'True' "
            "-preprocessor:__choice__ 'no_preprocessing' "
            "-regressor:__choice__ 'random_forest'",
            "R2_METRIC": '--initial-challengers " '
            "-imputation:strategy 'mean' "
            "-one_hot_encoding:minimum_fraction '0.01' "
            "-one_hot_encoding:use_minimum_fraction 'True' "
            "-preprocessor:__choice__ 'no_preprocessing' "
            "-regressor:__choice__ 'random_forest'",
        }

        for dataset_name, task, initial_challengers in [
            (dataset_name_regression, REGRESSION, initial_challengers_regression),
            (
                dataset_name_classification,
                MULTICLASS_CLASSIFICATION,
                initial_challengers_classification,
            ),
        ]:

            for metric in initial_challengers:
                configuration_space = get_configuration_space(
                    {"metric": metric, "task": task, "is_sparse": False},
                    include={"feature_preprocessor": ["no_preprocessing"]},
                )

                X_train, Y_train, X_test, Y_test = get_dataset(dataset_name)
                categorical = {i: False for i in range(X_train.shape[1])}

                meta_features_label = _calculate_metafeatures(
                    X_train, Y_train, categorical, dataset_name, task
                )
                meta_features_encoded_label = _calculate_metafeatures_encoded(
                    X_train, Y_train, categorical, dataset_name, task
                )

                initial_configuration_strings_for_smac = suggest_via_metalearning(
                    meta_features_label,
                    meta_features_encoded_label,
                    configuration_space,
                    dataset_name,
                    metric,
                    task,
                    False,
                    1,
                    None,
                )

                print(metric)
                print(initial_configuration_strings_for_smac[0])
                self.assertTrue(
                    initial_configuration_strings_for_smac[0].startswith(
                        initial_challengers[metric]
                    )
                )

    def test_metadata_directory(self):
        # Test that metadata directory is set correctly (if user specifies,
        # Auto-sklearn should check that the directory exists. If not, it
        # should use the default directory.
        dask_client = unittest.mock.Mock()
        automl1 = AutoSklearnClassifier(
            time_left_for_this_task=30,
            per_run_time_limit=5,
            metadata_directory="pyMetaLearn/metadata_dir",  # user metadata_dir
            dask_client=dask_client,
        )
        self.assertEqual(automl1.metadata_directory, "pyMetaLearn/metadata_dir")

        automl2 = AutoSklearnClassifier(  # default metadata_dir
            time_left_for_this_task=30,
            per_run_time_limit=5,
            dask_client=dask_client,
        )
        self.assertIsNone(automl2.metadata_directory)

        nonexistent_dir = "nonexistent_dir"
        automl3 = AutoSklearnClassifier(
            time_left_for_this_task=30,
            per_run_time_limit=5,
            metadata_directory=nonexistent_dir,  # user specified metadata_dir
            dask_client=dask_client,
            ensemble_class=None,
        )
        X, y = load_breast_cancer(return_X_y=True)
        self.assertRaisesRegex(
            ValueError,
            "The specified metadata directory "
            "'%s' does not exist!" % nonexistent_dir,
            automl3.fit,
            X=X,
            y=y,
        )



import sklearn.neighbors

from autosklearn.pipeline.components.classification.k_nearest_neighbors import (
    KNearestNeighborsClassifier,
)

from .test_base import BaseClassificationComponentTest


class KNearestNeighborsComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.959999999999999
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 1.381551055796429
    res["default_iris_sparse"] = 0.82
    res["default_digits"] = 0.93321190042501523
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.99574984820886459
    res["default_digits_multilabel"] = 0.93433756191199024
    res["default_digits_multilabel_proba"] = 0.9713841334968244

    sk_mod = sklearn.neighbors.KNeighborsClassifier
    module = KNearestNeighborsClassifier



import numpy as np
import pandas as pd

from autosklearn.pipeline.components.data_preprocessing.text_encoding.tfidf_encoding import (  # noqa: E501
    TfidfEncoder as Vectorizer,
)

import pytest


@pytest.mark.parametrize(
    "analyzer,per_column",
    [("word", True), ("word", False), ("char", True), ("char", False)],
)
def test_fit_transform(analyzer, per_column):
    X = pd.DataFrame(
        {
            "col1": ["hello world", "hello mars"],
            "col2": ["Test Test", "This is a test column"],
        }
    ).astype({"col1": "string", "col2": "string"})
    Vectorizer_fitted = Vectorizer(
        analyzer=analyzer,
        per_column=per_column,
        random_state=1,
    ).fit(X.copy())

    if per_column:
        for column in X.columns:
            Yt = Vectorizer_fitted.preprocessor[column].vocabulary_
            if column == "col1":
                if analyzer == "word":
                    words = sorted(
                        [
                            "hello",
                            "world",
                            "mars",
                        ]
                    )  # is ignored by TFIDFVectorizer
                    Y = {key: idx for idx, key in enumerate(words)}
                    assert Yt == Y
                else:
                    words = sorted(
                        [
                            "hell",
                            "ello",
                            "llo ",
                            "lo w",
                            "o wo",
                            " wor",
                            "worl",
                            "orld",
                            "lo m",
                            "o ma",
                            " mar",
                            "mars",
                        ]
                    )
                    Y = {key: idx for idx, key in enumerate(words)}
                    assert Yt == Y
            elif column == "col2":
                if analyzer == "word":
                    words = sorted(
                        [
                            "test",
                            "this",
                            "is",  # "a" is not added, len(...)=1,
                            "column",
                        ]
                    )  # is ignored by TFIDFVectorizer
                    Y = {key: idx for idx, key in enumerate(words)}
                    assert Yt == Y
                else:
                    words = sorted(
                        [
                            "test",
                            "est ",
                            "st t",
                            "t te",
                            " tes",
                            "this",
                            "his ",
                            "is i",
                            "s is",
                            " is ",
                            "is a",
                            "s a ",
                            " a t",
                            "a te",
                            "st c",
                            "t co",
                            " col",
                            "colu",
                            "olum",
                            "lumn",
                        ]
                    )
                    Y = {key: idx for idx, key in enumerate(words)}
                    assert Yt == Y
            else:
                raise ValueError(column)
    else:
        Yt = Vectorizer_fitted.preprocessor.vocabulary_
        if analyzer == "word":
            words = sorted(
                [
                    "hello",
                    "world",
                    "mars",
                    "test",
                    "this",
                    "is",  # "a" is not added, len(...)=1,
                    "column",
                ]
            )  # is ignored by TFIDFVectorizer
            Y = {key: idx for idx, key in enumerate(words)}
            assert Yt == Y
        else:
            words = sorted(
                [
                    "hell",
                    "ello",
                    "llo ",
                    "lo w",
                    "o wo",
                    " wor",
                    "worl",
                    "orld",
                    "lo m",
                    "o ma",
                    " mar",
                    "mars",
                    "test",
                    "est ",
                    "st t",
                    "t te",
                    " tes",
                    "this",
                    "his ",
                    "is i",
                    "s is",
                    " is ",
                    "is a",
                    "s a ",
                    " a t",
                    "a te",
                    "st c",
                    "t co",
                    " col",
                    "colu",
                    "olum",
                    "lumn",
                ]
            )
            Y = {key: idx for idx, key in enumerate(words)}
            assert Yt == Y


@pytest.mark.parametrize("per_column", [True, False])
def test_transform(per_column):
    X = pd.DataFrame(
        {
            "col1": ["hello world", "hello mars"],
            "col2": ["Test Test", "This is a test column"],
        }
    ).astype({"col1": "string", "col2": "string"})
    vectorizer = Vectorizer(
        per_column=per_column,
        analyzer="word",
        random_state=1,
    )
    X_t = vectorizer.fit_transform(X.copy())

    if per_column:
        # ['hello', 'mars', 'world', 'column', 'is', 'test', 'this']
        y = np.array(
            [
                [
                    0.57974,
                    0.0,
                    0.8148,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                ],
                [0.57974, 0.8148, 0.0, 0.53405, 0.53405, 0.37998, 0.53405],
            ]
        )
        np.testing.assert_almost_equal(X_t.toarray(), y, decimal=5)
    else:
        print(vectorizer.preprocessor.vocabulary_)
        # 'column', 'hello', 'is', 'mars', 'test', 'this', 'world
        y = np.array(
            [
                [0.0, 0.61913, 0.0, 0.0, 1.0, 0.0, 0.78529],
                [0.52547, 0.61913, 0.52547, 0.78529, 0.41429, 0.52547, 0.0],
            ]
        )
        np.testing.assert_almost_equal(X_t.toarray(), y, decimal=5)


def test_check_shape():
    X = pd.DataFrame(
        {
            "col1": ["hello world", "this is test"],
            "col2": ["test test", "test test"],
        }
    ).astype({"col1": "string", "col2": "string"})
    X_t = Vectorizer(
        per_column=True,
        analyzer="word",
        random_state=1,
    ).fit_transform(X.copy())

    assert X_t.shape == (2, 6)

    X_t = Vectorizer(
        analyzer="word",
        per_column=False,
        random_state=1,
    ).fit_transform(X.copy())

    assert X_t.shape == (2, 5)


def test_check_nan():
    X = pd.DataFrame(
        {
            "col1": ["hello world", "this is test", None],
            "col2": ["test test", "test test", "test"],
        }
    ).astype({"col1": "string", "col2": "string"})
    X_t = Vectorizer(
        per_column=True,
        analyzer="word",
        random_state=1,
    ).fit_transform(X.copy())
    assert X_t.shape == (3, 6)

    X_t = Vectorizer(
        analyzer="word",
        per_column=False,
        random_state=1,
    ).fit_transform(X.copy())
    assert X_t.shape == (3, 5)


def test_check_vocabulary():
    X = pd.DataFrame(
        {
            "col1": ["hello world", "this is test", None],
            "col2": ["test test", "test test", "test"],
        }
    ).astype({"col1": "string", "col2": "string"})
    vectorizer = Vectorizer(
        per_column=True,
        analyzer="word",
        random_state=1,
    ).fit(X.copy())
    assert vectorizer.preprocessor["col1"].vocabulary_ == {
        "hello": 0,
        "world": 4,
        "this": 3,
        "is": 1,
        "test": 2,
    }
    assert vectorizer.preprocessor["col2"].vocabulary_ == {"test": 0}

    vectorizer = Vectorizer(
        analyzer="word",
        per_column=False,
        random_state=1,
    ).fit(X.copy())
    assert vectorizer.preprocessor.vocabulary_ == {
        "hello": 0,
        "world": 4,
        "this": 3,
        "is": 1,
        "test": 2,
    }



import sklearn.ensemble

from autosklearn.pipeline.components.regression.random_forest import RandomForest

from .test_base import BaseRegressionComponentTest


class RandomForestComponentTest(BaseRegressionComponentTest):
    __test__ = True

    res = dict()
    res["default_boston"] = 0.8410063895401654
    res["boston_n_calls"] = 9
    res["default_boston_iterative"] = res["default_boston"]
    res["default_boston_sparse"] = 0.4194462097407078
    res["default_boston_iterative_sparse"] = res["default_boston_sparse"]
    res["default_diabetes"] = 0.3496051170409269
    res["diabetes_n_calls"] = 9
    res["default_diabetes_iterative"] = res["default_diabetes"]
    res["default_diabetes_sparse"] = 0.2383300978781976
    res["default_diabetes_iterative_sparse"] = res["default_diabetes_sparse"]

    sk_mod = sklearn.ensemble.RandomForestRegressor
    module = RandomForest
    step_hyperparameter = {
        "name": "n_estimators",
        "value": module.get_max_iter(),
    }



import sklearn.linear_model

from autosklearn.pipeline.components.regression.libsvm_svr import LibSVM_SVR

from .test_base import BaseRegressionComponentTest


class SupportVectorComponentTest(BaseRegressionComponentTest):
    __test__ = True

    res = dict()
    res["default_boston"] = -0.030006883949312613
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = -0.062749211736050192
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.12849591861430087
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = 0.0098877566961463881
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.svm.SVR

    module = LibSVM_SVR



from typing import Optional

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import (
    AutoSklearnClassificationAlgorithm,
    IterativeComponentWithSampleWeight,
)
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.pipeline.implementations.util import softmax
from autosklearn.util.common import check_for_bool


class PassiveAggressive(
    IterativeComponentWithSampleWeight,
    AutoSklearnClassificationAlgorithm,
):
    def __init__(self, C, fit_intercept, tol, loss, average, random_state=None):
        self.C = C
        self.fit_intercept = fit_intercept
        self.average = average
        self.tol = tol
        self.loss = loss
        self.random_state = random_state
        self.estimator = None
        self.max_iter = self.get_max_iter()
        self.n_iter_ = None

    @staticmethod
    def get_max_iter():
        return 1024

    def get_current_iter(self):
        return self.n_iter_

    def iterative_fit(self, X, y, n_iter=2, refit=False, sample_weight=None):
        from sklearn.linear_model import PassiveAggressiveClassifier

        # Need to fit at least two iterations, otherwise early stopping will not
        # work because we cannot determine whether the algorithm actually
        # converged. The only way of finding this out is if the sgd spends less
        # iterations than max_iter. If max_iter == 1, it has to spend at least
        # one iteration and will always spend at least one iteration, so we
        # cannot know about convergence.

        if refit:
            self.estimator = None
            self.n_iter_ = None

        if self.estimator is None:
            self.fully_fit_ = False

            self.average = check_for_bool(self.average)
            self.fit_intercept = check_for_bool(self.fit_intercept)
            self.tol = float(self.tol)
            self.C = float(self.C)

            call_fit = True
            self.estimator = PassiveAggressiveClassifier(
                C=self.C,
                fit_intercept=self.fit_intercept,
                max_iter=n_iter,
                tol=self.tol,
                loss=self.loss,
                shuffle=True,
                random_state=self.random_state,
                warm_start=True,
                average=self.average,
            )
            self.classes_ = np.unique(y.astype(int))
        else:
            call_fit = False

        # Fallback for multilabel classification
        if len(y.shape) > 1 and y.shape[1] > 1:
            import sklearn.multiclass

            self.estimator.max_iter = self.get_max_iter()
            self.estimator = sklearn.multiclass.OneVsRestClassifier(
                self.estimator, n_jobs=1
            )
            self.estimator.fit(X, y)
            self.fully_fit_ = True
        else:
            if call_fit:
                self.estimator.fit(X, y)
                self.n_iter_ = n_iter
            else:
                self.estimator.max_iter += n_iter
                self.estimator.max_iter = min(self.estimator.max_iter, self.max_iter)
                self.estimator._validate_params()
                lr = "pa1" if self.estimator.loss == "hinge" else "pa2"
                self.estimator._partial_fit(
                    X,
                    y,
                    alpha=1.0,
                    C=self.estimator.C,
                    loss="hinge",
                    learning_rate=lr,
                    max_iter=n_iter,
                    classes=None,
                    sample_weight=sample_weight,
                    coef_init=None,
                    intercept_init=None,
                )
                self.n_iter_ += self.estimator.n_iter_
                if (
                    self.estimator.max_iter >= self.max_iter
                    or self.estimator.max_iter > self.n_iter_
                ):
                    self.fully_fit_ = True

        return self

    def configuration_fully_fitted(self):
        if self.estimator is None:
            return False
        elif not hasattr(self, "fully_fit_"):
            return False
        else:
            return self.fully_fit_

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()

        df = self.estimator.decision_function(X)
        return softmax(df)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "PassiveAggressive Classifier",
            "name": "Passive Aggressive Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        C = UniformFloatHyperparameter("C", 1e-5, 10, 1.0, log=True)
        fit_intercept = UnParametrizedHyperparameter("fit_intercept", "True")
        loss = CategoricalHyperparameter(
            "loss", ["hinge", "squared_hinge"], default_value="hinge"
        )

        tol = UniformFloatHyperparameter(
            "tol", 1e-5, 1e-1, default_value=1e-4, log=True
        )
        # Note: Average could also be an Integer if > 1
        average = CategoricalHyperparameter(
            "average", ["False", "True"], default_value="False"
        )

        cs = ConfigurationSpace()
        cs.add_hyperparameters([loss, fit_intercept, tol, C, average])
        return cs



from __future__ import annotations

from typing import Callable

import pickle
import time
from pathlib import Path

import numpy as np

from autosklearn.ensemble_building import Run

from pytest_cases import parametrize


def test_is_dummy(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * We expect runs with an num_run (id) of 1 to be a dummy
    """
    run = make_run(id=1)
    assert run.is_dummy()

    run = make_run(id=2)
    assert not run.is_dummy()


def test_was_modified(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Should properly indicate when a file was modified
    """
    run = make_run()
    assert not run.was_modified()

    time.sleep(0.2)  # Just to give some time after creation
    path = run.pred_path("ensemble")
    path.touch()

    assert run.was_modified()


def test_record_modified_times_with_was_modified(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Updating the recorded times should not trigger `was_modified`
    * Should update the recorded times so `was_modified` will give False after being
      updated
    """
    run = make_run()
    path = run.pred_path("ensemble")

    time.sleep(0.2)
    run.record_modified_times()
    assert not run.was_modified()

    time.sleep(0.2)
    path.touch()
    assert run.was_modified()

    time.sleep(0.2)
    run.record_modified_times()
    assert not run.was_modified()


def test_predictions_pickled(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Should be able to load pickled predictions

    Note
    ----
    * Not sure this should be supported
    """
    run = make_run(predictions=[])
    x = np.array([0])

    path = run.pred_path("ensemble")
    with path.open("wb") as f:
        pickle.dump(x, f)

    assert run.predictions("ensemble") is not None


@parametrize(
    "precision, expected", [(16, np.float16), (32, np.float32), (64, np.float64)]
)
def test_predictions_precision(
    make_run: Callable[..., Run], precision: int, expected: type
) -> None:
    """
    Expects
    -------
    * Loading predictions with a given precision should load the expected type
    """
    run = make_run()
    assert run.predictions(precision=precision).dtype == expected


def test_caching(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Attempting to load the same predictions again will cause the result to be cached
    * Unloading the cache will cause it to reload and reread the predictions
    """
    run = make_run()

    assert len(run._cache) == 0
    first_load = run.predictions()  # Should cache result
    assert len(run._cache) == 1

    cache_load = run.predictions()  # Should use cache result
    assert len(run._cache) == 1

    # The should be the same object
    assert id(first_load) == id(cache_load)

    pickled_run = pickle.dumps(run)
    unpickled_run = pickle.loads(pickled_run)

    assert len(unpickled_run._cache) == 0
    unpickled_load = unpickled_run.predictions()  # Should have reloaded it
    assert len(unpickled_run._cache) == 1

    # Should not be the same object as before once pickled
    assert id(unpickled_load) != id(first_load)


def test_equality(make_run: Callable[..., Run]) -> None:
    """
    Expects
    -------
    * Two runs with the same id's should be considered equal
    * Otherwise, they should be considered different
    """
    r1 = make_run(id=1, budget=49.3, seed=3)
    r2 = make_run(id=1, budget=49.3, seed=3)

    assert r1 == r2

    r3 = make_run(id=1, budget=0.0, seed=3)

    assert r1 != r3
    assert r2 != r3


@parametrize(
    "name, expected",
    [
        ("0_0_0.0", True),
        ("1_152_64.24", True),
        ("123412_3462_100.0", True),
        ("tmp_sf12198", False),
        ("tmp_0_0_0.0", False),
    ],
)
def test_valid(name: str, expected: bool) -> None:
    """
    Expects
    -------
    * Should be able to correctly consider valid run dir names
    """
    path = Path(name)
    assert Run.valid(path) == expected



# -*- encoding: utf-8 -*-
import builtins

from autosklearn.automl_common.common.utils.backend import Backend

import unittest
import unittest.mock


class BackendModelsTest(unittest.TestCase):
    class BackendStub(Backend):
        def __init__(self):
            self.__class__ = Backend

    def setUp(self):
        self.backend = self.BackendStub()
        self.backend.internals_directory = "/"

    @unittest.mock.patch("pickle.load")
    @unittest.mock.patch("os.path.exists")
    def test_load_model_by_seed_and_id(self, exists_mock, pickleLoadMock):
        exists_mock.return_value = False
        open_mock = unittest.mock.mock_open(read_data="Data")
        with unittest.mock.patch(
            "autosklearn.automl_common.common.utils.backend.open",
            open_mock,
            create=True,
        ):
            seed = 13
            idx = 17
            budget = 50.0
            expected_model = self._setup_load_model_mocks(
                open_mock, pickleLoadMock, seed, idx, budget
            )

            actual_model = self.backend.load_model_by_seed_and_id_and_budget(
                seed, idx, budget
            )

            self.assertEqual(expected_model, actual_model)

    @unittest.mock.patch("pickle.load")
    @unittest.mock.patch.object(builtins, "open")
    @unittest.mock.patch("os.path.exists")
    def test_loads_models_by_identifiers(self, exists_mock, openMock, pickleLoadMock):
        exists_mock.return_value = True
        seed = 13
        idx = 17
        budget = 50.0
        expected_model = self._setup_load_model_mocks(
            openMock, pickleLoadMock, seed, idx, budget
        )
        expected_dict = {(seed, idx, budget): expected_model}

        actual_dict = self.backend.load_models_by_identifiers([(seed, idx, budget)])

        self.assertIsInstance(actual_dict, dict)
        self.assertDictEqual(expected_dict, actual_dict)

    def _setup_load_model_mocks(self, openMock, pickleLoadMock, seed, idx, budget):
        model_path = "/runs/%s_%s_%s/%s.%s.%s.model" % (
            seed,
            idx,
            budget,
            seed,
            idx,
            budget,
        )
        file_handler = "file_handler"
        expected_model = "model"

        fileMock = unittest.mock.MagicMock()
        fileMock.__enter__.return_value = file_handler

        openMock.side_effect = (
            lambda path, flag: fileMock if path == model_path and flag == "rb" else None
        )
        pickleLoadMock.side_effect = (
            lambda fh: expected_model if fh == file_handler else None
        )

        return expected_model
